@@ -1,6 +1,6 @@
 # pylint: disable=wildcard-import
 
-__version__ = (8, 19, 4, 5)
+__version__ = (8, 20, 5, 27)
 
 if bytes is str:
     raise Exception("""
@@ -13,7 +13,7 @@ If you want to continue using the most recent version of angr (you definitely
 want that, trust us) you should upgrade to python 3. It's like getting your
 vaccinations. It hurts a little bit initially but in the end it's worth it.
 
-For more information, see here: https://docs.angr.io/MIGRATION.html
+For more information, see here: https://docs.angr.io/appendix/migration
 
 Good luck!
 """)
@@ -56,11 +56,12 @@ from . import sim_type as types
 from .state_hierarchy import StateHierarchy
 
 from .sim_state import SimState
-from .engines import SimEngineVEX, SimEngine
+from . import engines
 from .calling_conventions import DEFAULT_CC, SYSCALL_CC, PointerWrapper, SimCC
 from .storage.file import SimFileBase, SimFile, SimPackets, SimFileStream, SimPacketsStream, SimFileDescriptor, SimFileDescriptorDuplex
 from .state_plugins.filesystem import SimMount, SimHostFilesystem
 from .state_plugins.heap import SimHeapBrk, SimHeapPTMalloc, PTChunk
+from .distributed import Server
 
 # for compatibility reasons
 from . import sim_manager as manager
