@@ -5,6 +5,8 @@ from . import MipsElfFastResolver
 from . import X86ElfPicPltResolver
 from . import JumpTableResolver
 from . import X86PeIatResolver
+from . import AMD64ElfGotResolver
+from . import ArmElfFastResolver
 
 
 DEFAULT_RESOLVERS = {
@@ -12,9 +14,24 @@ DEFAULT_RESOLVERS = {
         cle.MetaELF: [ X86ElfPicPltResolver, ],
         cle.PE: [ X86PeIatResolver, ],
     },
+    'AMD64': {
+        cle.MetaELF: [ AMD64ElfGotResolver, ],
+    },
     'MIPS32': {
         cle.MetaELF: [ MipsElfFastResolver, ],
     },
+    'MIPS64': {
+        cle.MetaELF: [ MipsElfFastResolver, ],
+    },
+    'ARMEL': {
+        cle.MetaELF: [ ArmElfFastResolver, ]
+    },
+    'ARMHF': {
+        cle.MetaELF: [ ArmElfFastResolver, ]
+    },
+    'ARMCortexM': {
+        cle.MetaELF: [ ArmElfFastResolver, ]
+    },
     'ALL': [ JumpTableResolver ],
 }
 
