@@ -1,39 +1,62 @@
+from typing import Optional, Iterable, Set, Union, TYPE_CHECKING
 import logging
-import struct
 
 import pyvex
 
-from .atoms import Register, MemoryLocation, Parameter
-from .constants import OP_BEFORE, OP_AFTER
-from .dataset import DataSet
-from .external_codeloc import ExternalCodeLocation
-from .undefined import Undefined
-from ...engines.light import SimEngineLightVEX, SpOffset
-from ...engines.vex.irop import operations as vex_operations
+from ...engines.light import SimEngineLight, SimEngineLightVEXMixin, SpOffset, RegisterOffset
+from ...engines.vex.claripy.irop import operations as vex_operations
 from ...errors import SimEngineError
+from ...calling_conventions import DEFAULT_CC, SimRegArg, SimStackArg
+from ...knowledge_plugins.key_definitions.definition import Definition
+from ...knowledge_plugins.key_definitions.atoms import Register, MemoryLocation, Parameter, Tmp
+from ...knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
+from ...knowledge_plugins.key_definitions.dataset import DataSet
+from ...knowledge_plugins.key_definitions.undefined import Undefined, undefined
+from .rd_state import ReachingDefinitionsState
+from .external_codeloc import ExternalCodeLocation
+
+if TYPE_CHECKING:
+    from ...calling_conventions import SimCC
+    from ...knowledge_plugins import FunctionManager
+
 
 l = logging.getLogger(name=__name__)
 
 
-class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
-    def __init__(self, project, current_local_call_depth, maximum_local_call_depth, function_handler=None):
+class SimEngineRDVEX(
+    SimEngineLightVEXMixin,
+    SimEngineLight,
+):  # pylint:disable=abstract-method
+    def __init__(self, project, current_local_call_depth, maximum_local_call_depth, functions=None,
+                 function_handler=None):
         super(SimEngineRDVEX, self).__init__()
         self.project = project
+        self.functions: Optional['FunctionManager'] = functions
         self._current_local_call_depth = current_local_call_depth
         self._maximum_local_call_depth = maximum_local_call_depth
         self._function_handler = function_handler
+        self._visited_blocks = None
+        self._dep_graph = None
+
+        self.state: ReachingDefinitionsState
 
     def process(self, state, *args, **kwargs):
+        self._dep_graph = kwargs.pop('dep_graph', None)
+        self._visited_blocks = kwargs.pop('visited_blocks', None)
+
         # we are using a completely different state. Therefore, we directly call our _process() method before
         # SimEngine becomes flexible enough.
         try:
-            self._process(state, None, block=kwargs.pop('block', None))
+            self._process(
+                state,
+                None,
+                block=kwargs.pop('block', None),
+            )
         except SimEngineError as e:
             if kwargs.pop('fail_fast', False) is True:
                 raise e
-            else:
-                l.error(e)
-        return self.state
+            l.error(e)
+        return self.state, self._visited_blocks
 
     #
     # Private methods
@@ -57,10 +80,22 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
         if self.state.analysis:
             self.state.analysis.insn_observe(self.ins_addr, stmt, self.block, self.state, OP_AFTER)
 
+    def _handle_WrTmp(self, stmt: pyvex.IRStmt.WrTmp):
+        super()._handle_WrTmp(stmt)
+        self.state.kill_and_add_definition(Tmp(stmt.tmp, self.tyenv.sizeof(stmt.tmp) // 8),
+                                           self._codeloc(),
+                                           self.tmps[stmt.tmp])
+
+    def _handle_WrTmpData(self, tmp: int, data):
+        super()._handle_WrTmpData(tmp, data)
+        self.state.kill_and_add_definition(Tmp(tmp, self.tyenv.sizeof(tmp)),
+                                           self._codeloc(),
+                                           self.tmps[tmp])
+
     # e.g. PUT(rsp) = t2, t2 might include multiple values
     def _handle_Put(self, stmt):
-        reg_offset = stmt.offset
-        size = stmt.data.result_size(self.tyenv) // 8
+        reg_offset: int = stmt.offset
+        size: int = stmt.data.result_size(self.tyenv) // 8
         reg = Register(reg_offset, size)
         data = self._expr(stmt.data)
 
@@ -68,16 +103,45 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             l.info('Data to write into register <%s> with offset %d undefined, ins_addr = %#x.',
                    self.arch.register_names[reg_offset], reg_offset, self.ins_addr)
 
+        # special handling for references to stack variables
+        for d in data:
+            if isinstance(d, SpOffset) and isinstance(d.offset, int):
+                self.state.add_use(MemoryLocation(d, 1), self._codeloc())
+
         self.state.kill_and_add_definition(reg, self._codeloc(), data)
 
     # e.g. STle(t6) = t21, t6 and/or t21 might include multiple values
-    # sync with _handle_StoreG()
     def _handle_Store(self, stmt):
-
         addr = self._expr(stmt.addr)
         size = stmt.data.result_size(self.tyenv) // 8
         data = self._expr(stmt.data)
 
+        self._store_core(addr, size, data)
+
+    def _handle_StoreG(self, stmt: pyvex.IRStmt.StoreG):
+        guard = self._expr(stmt.guard)
+        if guard.data == {True}:
+            addr = self._expr(stmt.addr)
+            size = stmt.data.result_size(self.tyenv) // 8
+            data = self._expr(stmt.data)
+
+            self._store_core(addr, size, data)
+        elif guard.data == {False}:
+            pass
+        else:
+            # guard.data == {True, False}
+            # get current data
+            addr = self._expr(stmt.addr)
+            size = stmt.data.result_size(self.tyenv) // 8
+            data_old = self._load_core(addr, size, stmt.endness)
+            data = self._expr(stmt.data)
+
+            self._store_core(addr, size, data, data_old=data_old)
+
+    def _store_core(self, addr: Iterable[Union[int,SpOffset,Undefined]], size: int, data, data_old=None):
+        if data_old:
+            data.update(data_old)
+
         for a in addr:
             if type(a) is Undefined:
                 l.info('Memory address undefined, ins_addr = %#x.', self.ins_addr)
@@ -85,50 +149,14 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
                 if any(type(d) is Undefined for d in data):
                     l.info('Data to write at address %#x undefined, ins_addr = %#x.', a, self.ins_addr)
 
-                memloc = MemoryLocation(a, size)
-                # different addresses are not killed by a subsequent iteration, because kill only removes entries
-                # with same index and same size
-                self.state.kill_and_add_definition(memloc, self._codeloc(), data)
-
-    # sync with _handle_Store()
-    def _handle_StoreG(self, stmt):
-        guard = self._expr(stmt.guard)
-        if guard.data == {True}:
-            self._handle_Store(stmt)
-        elif guard.data == {False}:
-            pass
-        else:
-            # guard.data == {True, False}
-            addr = self._expr(stmt.addr)
-            size = stmt.data.result_size(self.tyenv) // 8
-
-            # get current data
-            load_end = stmt.end
-            load_ty = self.tyenv.lookup(stmt.data.tmp)
-            load_addr = stmt.addr
-            load_expr = pyvex.IRExpr.Load(load_end, load_ty, load_addr)
-            data_old = self._handle_Load(load_expr)
-
-            # get new data
-            data_new = self._expr(stmt.data)
-
-            # merge old and new data
-            data_new.update(data_old)
-
-            for a in addr:
-                if type(a) is Undefined:
-                    l.info('Memory address undefined, ins_addr = %#x.', self.ins_addr)
-                else:
-                    if any(type(d) is Undefined for d in data_new):
-                        l.info('Data to write at address %#x undefined, ins_addr = %#x.', a, self.ins_addr)
-
+                if isinstance(a, int) or (isinstance(a, SpOffset) and isinstance(a.offset, int)):
                     memloc = MemoryLocation(a, size)
                     # different addresses are not killed by a subsequent iteration, because kill only removes entries
                     # with same index and same size
-                    self.state.kill_and_add_definition(memloc, self._codeloc(), data_new)
+                    self.state.kill_and_add_definition(memloc, self._codeloc(), data)
 
     def _handle_LoadG(self, stmt):
-        guard = self._expr(stmt.guard)
+        guard: DataSet = self._expr(stmt.guard)
         if guard.data == {True}:
             # FIXME: full conversion support
             if stmt.cvt.find('Ident') < 0:
@@ -149,7 +177,9 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             self._handle_WrTmpData(stmt.dst, DataSet(data, load_expr.result_size(self.tyenv)))
 
     def _handle_Exit(self, stmt):
-        pass
+        guard = self._expr(stmt.guard)
+        target = stmt.dst.value
+        self.state.mark_guard(self._codeloc(), guard, target)
 
     def _handle_IMark(self, stmt):
         pass
@@ -157,32 +187,65 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
     def _handle_AbiHint(self, stmt):
         pass
 
+    def _handle_LLSC(self, stmt: pyvex.IRStmt.LLSC):
+        if stmt.storedata is None:
+            # load-link
+            addr = self._expr(stmt.addr)
+            size = self.tyenv.sizeof(stmt.result) // self.arch.byte_width
+            load_result = self._load_core(addr, size, stmt.endness)
+            self.tmps[stmt.result] = load_result
+            self.state.kill_and_add_definition(Tmp(stmt.result, self.tyenv.sizeof(stmt.result) // self.arch.byte_width),
+                                               self._codeloc(),
+                                               load_result)
+        else:
+            # store-conditional
+            storedata = self._expr(stmt.storedata)
+            addr = self._expr(stmt.addr)
+            size = self.tyenv.sizeof(stmt.storedata.tmp) // self.arch.byte_width
+
+            self._store_core(addr, size, storedata)
+            self.tmps[stmt.result] = DataSet({1}, 1)
+            self.state.kill_and_add_definition(Tmp(stmt.result, self.tyenv.sizeof(stmt.result) // self.arch.byte_width),
+                                               self._codeloc(),
+                                               self.tmps[stmt.result])
+
     #
     # VEX expression handlers
     #
 
-    def _handle_RdTmp(self, expr):
-        tmp = expr.tmp
+    def _expr(self, expr) -> DataSet:
+        data = super()._expr(expr)
+        if data is None:
+            bits = expr.result_size(self.tyenv)
+            data = DataSet(undefined, bits)
+        return data
+
+    def _handle_RdTmp(self, expr: pyvex.IRExpr.RdTmp) -> Optional[DataSet]:
+        tmp: int = expr.tmp
+
+        self.state.add_use(Tmp(tmp, expr.result_size(self.tyenv) // self.arch.byte_width), self._codeloc())
 
         if tmp in self.tmps:
             return self.tmps[tmp]
-        bits = expr.result_size(self.tyenv)
-        return DataSet(Undefined(bits), bits)
+        return None
 
     # e.g. t0 = GET:I64(rsp), rsp might be defined multiple times
-    def _handle_Get(self, expr):
+    def _handle_Get(self, expr: pyvex.IRExpr.Get) -> Optional[DataSet]:
 
-        reg_offset = expr.offset
-        bits = expr.result_size(self.tyenv)
-        size = bits // self.arch.byte_width
+        reg_offset: int = expr.offset
+        bits: int = expr.result_size(self.tyenv)
+        size: int = bits // self.arch.byte_width
 
         # FIXME: size, overlapping
-        data = set()
-        current_defs = self.state.register_definitions.get_objects_by_offset(reg_offset)
+        data: Set[Union[Undefined,RegisterOffset,int]] = set()
+        current_defs: Iterable[Definition] = self.state.register_definitions.get_objects_by_offset(reg_offset)
         for current_def in current_defs:
             data.update(current_def.data)
         if len(data) == 0:
-            data.add(Undefined(bits))
+            # no defs can be found. add a fake definition
+            data.add(undefined)
+            self.state.kill_and_add_definition(Register(reg_offset, size), self._external_codeloc(),
+                                               DataSet(data, bits))
         if any(type(d) is Undefined for d in data):
             l.info('Data in register <%s> with offset %d undefined, ins_addr = %#x.',
                    self.arch.register_names[reg_offset], reg_offset, self.ins_addr)
@@ -198,9 +261,16 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
         bits = expr.result_size(self.tyenv)
         size = bits // self.arch.byte_width
 
+        return self._load_core(addr, size, expr.endness)
+
+    def _load_core(self, addr: Iterable[Union[int,SpOffset]], size: int, endness: str):  # pylint:disable=unused-argument
+
+        current_defs: Iterable[Definition]
+
         data = set()
         for a in addr:
             if isinstance(a, int):
+                # Load data from a global region
                 current_defs = self.state.memory_definitions.get_objects_by_offset(a)
                 if current_defs:
                     for current_def in current_defs:
@@ -209,19 +279,30 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
                         l.info('Memory at address %#x undefined, ins_addr = %#x.', a, self.ins_addr)
                 else:
                     try:
-                        data.add(self.state.loader.memory.unpack_word(a, size=size))
+                        data.add(self.project.loader.memory.unpack_word(a, size=size))
                     except KeyError:
                         pass
 
                 # FIXME: _add_memory_use() iterates over the same loop
-                self.state.add_use(MemoryLocation(a, size), self._codeloc())
+                memory_location = MemoryLocation(a, size)
+                self.state.add_use(memory_location, self._codeloc())
+            elif isinstance(a, SpOffset) and isinstance(a.offset, int):
+                # Load data from a local variable
+                current_defs = self.state.stack_definitions.get_objects_by_offset(a.offset)
+                if current_defs:
+                    for def_ in current_defs:
+                        data.update(def_.data)
+                else:
+                    data.add(undefined)
+                memory_location = MemoryLocation(a, size)
+                self.state.add_use(memory_location, self._codeloc())
             else:
                 l.info('Memory address undefined, ins_addr = %#x.', self.ins_addr)
 
         if len(data) == 0:
-            data.add(Undefined(bits))
+            data.add(undefined)
 
-        return DataSet(data, bits)
+        return DataSet(data, size * self.arch.byte_width)
 
     # CAUTION: experimental
     def _handle_ITE(self, expr):
@@ -281,7 +362,7 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             if isinstance(e0, int):
                 return DataSet(e0 != 1, expr.result_size(self.tyenv))
 
-        l.warning('Comparison of multiple values / different types.')
+        # l.warning('Comparison of multiple values / different types.')
         return DataSet({True, False}, expr.result_size(self.tyenv))
 
     def _handle_Not(self, expr):
@@ -293,7 +374,7 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             if isinstance(e0, int):
                 return DataSet(e0 == 0, expr.result_size(self.tyenv))
 
-        l.warning('Comparison of multiple values / different types.')
+        # l.warning('Comparison of multiple values / different types.')
         return DataSet({True, False}, expr.result_size(self.tyenv))
 
     #
@@ -316,7 +397,7 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
                         head = ((1 << e1) - 1) << (bits - e1)
                     data.add(head | (e0 >> e1))
                 except (ValueError, TypeError) as e:
-                    data.add(Undefined(bits))
+                    data.add(undefined)
                     l.warning(e)
 
         return DataSet(data, expr.result_size(self.tyenv))
@@ -332,7 +413,7 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             if isinstance(e0, int) and isinstance(e1, int):
                 return DataSet(e0 == e1, expr.result_size(self.tyenv))
 
-        l.warning('Comparison of multiple values / different types.')
+        # l.warning('Comparison of multiple values / different types.')
         return DataSet({True, False}, expr.result_size(self.tyenv))
 
     def _handle_CmpNE(self, expr):
@@ -346,7 +427,7 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             if isinstance(e0, int) and isinstance(e1, int):
                 return DataSet(e0 != e1, expr.result_size(self.tyenv))
 
-        l.warning('Comparison of multiple values / different types.')
+        # l.warning('Comparison of multiple values / different types.')
         return DataSet({True, False}, expr.result_size(self.tyenv))
 
     def _handle_CmpLT(self, expr):
@@ -360,7 +441,7 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
             if isinstance(e0, int) and isinstance(e1, int):
                 return DataSet(e0 < e1, expr.result_size(self.tyenv))
 
-        l.warning('Comparison of multiple values / different types.')
+        # l.warning('Comparison of multiple values / different types.')
         return DataSet({True, False}, expr.result_size(self.tyenv))
 
     # ppc only
@@ -380,58 +461,72 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
                 else:
                     return DataSet(0x02, expr.result_size(self.tyenv))
 
-        l.warning('Comparison of multiple values / different types.')
+        # l.warning('Comparison of multiple values / different types.')
         return DataSet({True, False}, expr.result_size(self.tyenv))
 
     def _handle_CCall(self, expr):
         bits = expr.result_size(self.tyenv)
-        return DataSet(Undefined(bits), bits)
+        for arg_expr in expr.args:
+            self._expr(arg_expr)
+        return DataSet(undefined, bits)
 
     #
     # User defined high level statement handlers
     #
 
-    def _handle_function(self):
+    def _handle_function(self, func_addr: Optional[DataSet], **kwargs):
+        skip_cc = self._handle_function_core(func_addr, **kwargs)
+        if not skip_cc:
+            self._handle_function_cc(func_addr)
+
+    def _handle_function_core(self, func_addr: Optional[DataSet], **kwargs) -> bool:  # pylint:disable=unused-argument
+
         if self._current_local_call_depth > self._maximum_local_call_depth:
             l.warning('The analysis reached its maximum recursion depth.')
-            return None
+            return False
 
-        defs_ip = self.state.register_definitions.get_objects_by_offset(self.arch.ip_offset)
-        if len(defs_ip) != 1:
-            l.error('Invalid definition(s) for IP.')
-            return None
+        if func_addr is None:
+            l.warning('Invalid type %s for IP.', type(func_addr).__name__)
+            handler_name = 'handle_unknown_call'
+            if hasattr(self._function_handler, handler_name):
+                executed_rda, state = getattr(self._function_handler, handler_name)(self.state, self._codeloc())
+                state: ReachingDefinitionsState
+                self.state = state
+            else:
+                l.warning('Please implement the unknown function handler with your own logic.')
+            return False
 
-        ip_data = next(iter(defs_ip)).data
-        if len(ip_data) != 1:
+        if len(func_addr) != 1:
+            # indirect call
             handler_name = 'handle_indirect_call'
             if hasattr(self._function_handler, handler_name):
                 _, state = getattr(self._function_handler, handler_name)(self.state, self._codeloc())
                 self.state = state
             else:
-                l.warning('Please implement the inderect function handler with your own logic.')
-            return None
+                l.warning('Please implement the indirect function handler with your own logic.')
+            return False
 
-        ip_addr = ip_data.get_first_element()
-        if not isinstance(ip_addr, int):
-            l.warning('Invalid type %s for IP.', type(ip_addr).__name__)
+        func_addr_int = func_addr.get_first_element()
+        if not isinstance(func_addr_int, int):
+            l.warning('Invalid type %s for IP.', type(func_addr_int).__name__)
             handler_name = 'handle_unknown_call'
             if hasattr(self._function_handler, handler_name):
                 executed_rda, state = getattr(self._function_handler, handler_name)(self.state, self._codeloc())
+                state: ReachingDefinitionsState
                 self.state = state
             else:
                 l.warning('Please implement the unknown function handler with your own logic.')
-            return None
+            return False
 
-        is_internal = False
+        # direct calls
         ext_func_name = None
-        if self.state.loader.main_object.contains_addr(ip_addr) is True:
-            ext_func_name = self.state.loader.find_plt_stub_name(ip_addr)
-            if ext_func_name is None:
-                is_internal = True
+        if self.project.loader.main_object.contains_addr(func_addr_int):
+            ext_func_name = self.project.loader.find_plt_stub_name(func_addr_int)
         else:
-            symbol = self.state.loader.find_symbol(ip_addr)
+            symbol = self.project.loader.find_symbol(func_addr_int)
             if symbol is not None:
                 ext_func_name = symbol.name
+        is_internal = ext_func_name is None
 
         executed_rda = False
         if ext_func_name is not None:
@@ -440,7 +535,8 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
                 executed_rda, state = getattr(self._function_handler, handler_name)(self.state, self._codeloc())
                 self.state = state
             else:
-                l.warning('Please implement the external function handler for %s() with your own logic.', ext_func_name)
+                l.warning('Please implement the external function handler for %s() with your own logic.',
+                          ext_func_name)
                 handler_name = 'handle_external_function_fallback'
                 if hasattr(self._function_handler, handler_name):
                     executed_rda, state = getattr(self._function_handler, handler_name)(self.state, self._codeloc())
@@ -448,30 +544,73 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
         elif is_internal is True:
             handler_name = 'handle_local_function'
             if hasattr(self._function_handler, handler_name):
-                executed_rda, state = getattr(self._function_handler, handler_name)(self.state,
-                                                                                    ip_addr,
-                                                                                    self._current_local_call_depth + 1,
-                                                                                    self._maximum_local_call_depth,
-                                                                                    self._codeloc(),
-                                                                                    )
+                executed_rda, state = getattr(self._function_handler, handler_name)(
+                    self.state,
+                    func_addr_int,
+                    self._current_local_call_depth + 1,
+                    self._maximum_local_call_depth,
+                    self._codeloc(),
+                )
                 self.state = state
             else:
-                l.warning('Please implement the local function handler with your own logic.')
+                # l.warning('Please implement the local function handler with your own logic.')
+                pass
         else:
-            l.warning('Could not find function name for external function at address %#x.', ip_addr)
+            l.warning('Could not find function name for external function at address %#x.', func_addr_int)
             handler_name = 'handle_unknown_call'
             if hasattr(self._function_handler, handler_name):
                 executed_rda, state = getattr(self._function_handler, handler_name)(self.state, self._codeloc())
                 self.state = state
             else:
                 l.warning('Please implement the unknown function handler with your own logic.')
+        skip_cc = executed_rda
 
-        # pop return address if necessary
-        if executed_rda is False and self.arch.call_pushes_ret is True:
+        return skip_cc
+
+    def _handle_function_cc(self, func_addr: Optional[DataSet]):  # pylint:disable=unused-argument
+
+        cc: Optional['SimCC'] = None
+        if func_addr is not None and len(func_addr) == 1 and self.functions is not None:
+            func_addr_int = next(iter(func_addr))
+            if self.functions.contains_addr(func_addr_int):
+                cc = self.functions[func_addr_int].calling_convention
+
+        if cc is None:
+            cc = DEFAULT_CC.get(self.arch.name, None)(self.arch)
+
+        if cc is not None:
+            # follow the calling convention and:
+            # - add uses for arguments
+            # - kill return value registers
+            # - caller-saving registers
+            if cc.args:
+                code_loc = self._codeloc()
+                for arg in cc.args:
+                    if isinstance(arg, SimRegArg):
+                        reg_offset, reg_size = self.arch.registers[arg.reg_name]
+                        self.state.add_use(Register(reg_offset, reg_size), code_loc)
+                    elif isinstance(arg, SimStackArg):
+                        self.state.add_use(MemoryLocation(SpOffset(self.arch.bits,
+                                                                   arg.stack_offset),
+                                                          arg.size * self.arch.byte_width),
+                                           code_loc)
+            if cc.RETURN_VAL is not None:
+                if isinstance(cc.RETURN_VAL, SimRegArg):
+                    reg_offset, reg_size = self.arch.registers[cc.RETURN_VAL.reg_name]
+                    atom = Register(reg_offset, reg_size)
+                    self.state.kill_and_add_definition(atom, self._codeloc(), DataSet({undefined}, reg_size * 8))
+            if cc.CALLER_SAVED_REGS is not None:
+                for reg in cc.CALLER_SAVED_REGS:
+                    reg_offset, reg_size = self.arch.registers[reg]
+                    atom = Register(reg_offset, reg_size)
+                    self.state.kill_and_add_definition(atom, self._codeloc(), DataSet({undefined}, reg_size * 8))
+
+        if self.arch.call_pushes_ret is True:
+            # pop return address if necessary
             defs_sp = self.state.register_definitions.get_objects_by_offset(self.arch.sp_offset)
             if len(defs_sp) == 0:
                 raise ValueError('No definition for SP found')
-            elif len(defs_sp) == 1:
+            if len(defs_sp) == 1:
                 sp_data = next(iter(defs_sp)).data.data
             else:  # len(defs_sp) > 1
                 sp_data = set()
@@ -482,11 +621,12 @@ class SimEngineRDVEX(SimEngineLightVEX):  # pylint:disable=abstract-method
                 raise ValueError('Invalid number of values for stack pointer.')
 
             sp_addr = next(iter(sp_data))
-            if not isinstance(sp_addr, int):
+            if isinstance(sp_addr, (int, SpOffset)):
+                sp_addr -= self.arch.stack_change
+            elif isinstance(sp_addr, Undefined):
+                pass
+            else:
                 raise TypeError('Invalid type %s for stack pointer.' % type(sp_addr).__name__)
 
             atom = Register(self.arch.sp_offset, self.arch.bytes)
-            sp_addr -= self.arch.stack_change
             self.state.kill_and_add_definition(atom, self._codeloc(), DataSet(sp_addr, self.arch.bits))
-
-        return None
