@@ -1,22 +1,31 @@
 import logging
+from typing import List
+
+from pyvex import IRSB
+
 l = logging.getLogger(name=__name__)
 
 import pyvex
 from archinfo import ArchARM
-from .engines import SimEngineVEX
 
-DEFAULT_VEX_ENGINE = SimEngineVEX(None)  # this is only used when Block is not initialized with a project
+from .protos import primitives_pb2 as pb2
+from .serializable import Serializable
+from .engines.vex import VEXLifter
+
+DEFAULT_VEX_ENGINE = VEXLifter(None)  # this is only used when Block is not initialized with a project
 
 
-class Block(object):
+class Block(Serializable):
     BLOCK_MAX_SIZE = 4096
 
     __slots__ = ['_project', '_bytes', '_vex', 'thumb', '_capstone', 'addr', 'size', 'arch', '_instructions',
-                 '_instruction_addrs', '_opt_level', '_vex_nostmt', '_collect_data_refs'
+                 '_instruction_addrs', '_opt_level', '_vex_nostmt', '_collect_data_refs', '_strict_block_end',
+                 '_cross_insn_opt',
                  ]
 
     def __init__(self, addr, project=None, arch=None, size=None, byte_string=None, vex=None, thumb=False, backup_state=None,
-                 extra_stop_points=None, opt_level=None, num_inst=None, traceflags=0, strict_block_end=None, collect_data_refs=False):
+                 extra_stop_points=None, opt_level=None, num_inst=None, traceflags=0, strict_block_end=None,
+                 collect_data_refs=False, cross_insn_opt=True):
 
         # set up arch
         if project is not None:
@@ -49,7 +58,7 @@ class Block(object):
             elif vex is not None:
                 size = vex.size
             else:
-                vex = self._vex_engine.lift(
+                vex = self._vex_engine.lift_vex(
                         clemory=project.loader.memory,
                         state=backup_state,
                         insn_bytes=byte_string,
@@ -61,6 +70,7 @@ class Block(object):
                         traceflags=traceflags,
                         strict_block_end=strict_block_end,
                         collect_data_refs=collect_data_refs,
+                        cross_insn_opt=cross_insn_opt,
                 )
                 size = vex.size
 
@@ -69,9 +79,11 @@ class Block(object):
         self._capstone = None
         self.size = size
         self._collect_data_refs = collect_data_refs
+        self._strict_block_end = strict_block_end
+        self._cross_insn_opt = cross_insn_opt
 
         self._instructions = num_inst
-        self._instruction_addrs = []
+        self._instruction_addrs = [] # type: List[int]
 
         self._parse_vex_info()
 
@@ -115,6 +127,7 @@ class Block(object):
     def __setstate__(self, data):
         for k, v in data.items():
             setattr(self, k, v)
+        self._capstone = None
 
     def __hash__(self):
         return hash((type(self), self.addr, self.bytes))
@@ -138,9 +151,9 @@ class Block(object):
             return self._project.factory.default_engine
 
     @property
-    def vex(self):
+    def vex(self) -> IRSB:
         if not self._vex:
-            self._vex = self._vex_engine.lift(
+            self._vex = self._vex_engine.lift_vex(
                     clemory=self._project.loader.memory if self._project is not None else None,
                     insn_bytes=self._bytes,
                     addr=self.addr,
@@ -150,6 +163,8 @@ class Block(object):
                     opt_level=self._opt_level,
                     arch=self.arch,
                     collect_data_refs=self._collect_data_refs,
+                    strict_block_end=self._strict_block_end,
+                    cross_insn_opt=self._cross_insn_opt,
             )
             self._parse_vex_info()
 
@@ -163,7 +178,7 @@ class Block(object):
         if self._vex:
             return self._vex
 
-        self._vex_nostmt = self._vex_engine.lift(
+        self._vex_nostmt = self._vex_engine.lift_vex(
             clemory=self._project.loader.memory if self._project is not None else None,
             insn_bytes=self._bytes,
             addr=self.addr,
@@ -174,6 +189,8 @@ class Block(object):
             arch=self.arch,
             skip_stmts=True,
             collect_data_refs=self._collect_data_refs,
+            strict_block_end=self._strict_block_end,
+            cross_insn_opt=self._cross_insn_opt,
         )
         return self._vex_nostmt
 
@@ -221,6 +238,26 @@ class Block(object):
 
         return self._instruction_addrs
 
+    @classmethod
+    def _get_cmsg(cls):
+        return pb2.Block()
+
+    def serialize_to_cmessage(self):
+        obj = self._get_cmsg()
+        obj.ea = self.addr
+        obj.size = self.size
+        obj.bytes = self.bytes
+
+        return obj
+
+    @classmethod
+    def parse_from_cmessage(cls, cmsg):
+        obj = cls(cmsg.ea,
+                  size=cmsg.size,
+                  byte_string=cmsg.bytes,
+                  )
+        return obj
+
 
 class SootBlock:
     def __init__(self, addr, project=None, arch=None):
@@ -239,7 +276,7 @@ class SootBlock:
 
     @property
     def soot(self):
-        return self._soot_engine.lift(self.addr, the_binary=self._the_binary)
+        return self._soot_engine.lift_soot(self.addr, the_binary=self._the_binary)
 
     @property
     def size(self):
@@ -254,7 +291,7 @@ class SootBlock:
         return SootBlockNode(self.addr, stmts_len, stmts=stmts)
 
 
-class CapstoneBlock(object):
+class CapstoneBlock:
     """
     Deep copy of the capstone blocks, which have serious issues with having extended lifespans
     outside of capstone itself
@@ -277,7 +314,7 @@ class CapstoneBlock(object):
         return '<CapstoneBlock for %#x>' % self.addr
 
 
-class CapstoneInsn(object):
+class CapstoneInsn:
     def __init__(self, capstone_insn):
         self.insn = capstone_insn
 
