@@ -5,7 +5,7 @@ from archinfo.arch_soot import ArchSoot
 import logging
 l = logging.getLogger(name=__name__)
 
-class SimSuccessors(object):
+class SimSuccessors:
     """
     This class serves as a categorization of all the kinds of result states that can come from a
     SimEngine run.
@@ -28,7 +28,8 @@ class SimSuccessors(object):
     :ivar unconstrained_successors:
                             Any state for which during the flattening process we find too many solutions.
 
-    A more detailed description of the successor lists may be found here: https://docs.angr.io/docs/simuvex.html
+    A more detailed description of the successor lists may be found here:
+    https://docs.angr.io/core-concepts/simulation#simsuccessors
     """
 
     def __init__(self, addr, initial_state):
@@ -183,7 +184,7 @@ class SimSuccessors(object):
                     ret_addr = state.mem[state.regs._sp].long.concrete
                 else:
                     ret_addr = state.solver.eval(state.regs._lr)
-            except SimSolverModeError:
+            except (SimSolverModeError, SimUnsatError):
                 # Use the address for UnresolvableCallTarget instead.
                 ret_addr = state.project.simos.unresolvable_call_target
 
@@ -194,7 +195,7 @@ class SimSuccessors(object):
 
             try:
                 stack_ptr = state.solver.eval(state.regs._sp)
-            except SimSolverModeError:
+            except (SimSolverModeError, SimUnsatError):
                 stack_ptr = 0
 
             new_frame = CallStack(
@@ -207,7 +208,10 @@ class SimSuccessors(object):
 
             state._inspect('call', BP_AFTER)
         else:
-            while state.solver.is_true(state.regs._sp > state.callstack.top.stack_ptr):
+            while True:
+                cur_sp = state.solver.max(state.regs._sp) if state.has_plugin('symbolizer') else state.regs._sp
+                if not state.solver.is_true(cur_sp > state.callstack.top.stack_ptr):
+                    break
                 state._inspect('return', BP_BEFORE, function_address=state.callstack.top.func_addr)
                 state.callstack.pop()
                 state._inspect('return', BP_AFTER)
@@ -284,7 +288,7 @@ class SimSuccessors(object):
                     # up, and create a "unknown syscall" stub for it.
                     self._fix_syscall_ip(state)
                     self.flat_successors.append(state)
-            except AngrUnsupportedSyscallError:
+            except (AngrUnsupportedSyscallError, AngrSyscallError):
                 self.unsat_successors.append(state)
 
         else:
@@ -292,10 +296,12 @@ class SimSuccessors(object):
             _max_targets = state.options.symbolic_ip_max_targets
             _max_jumptable_targets = state.options.jumptable_symbolic_ip_max_targets
             try:
+                skip_max_targets_warning = False
                 if o.NO_IP_CONCRETIZATION in state.options:
                     # Don't try to concretize the IP
                     cond_and_targets = [ (claripy.true, target) ]
                     max_targets = 0
+                    skip_max_targets_warning = True  # don't warn
                 elif o.KEEP_IP_SYMBOLIC in state.options:
                     s = claripy.Solver()
                     addrs = s.eval(target, _max_targets + 1, extra_constraints=tuple(state.ip_constraints))
@@ -316,11 +322,12 @@ class SimSuccessors(object):
                         max_targets = _max_jumptable_targets
 
                 if len(cond_and_targets) > max_targets:
-                    l.warning(
-                        "Exit state has over %d possible solutions. Likely unconstrained; skipping. %s",
-                        max_targets,
-                        target.shallow_repr()
-                    )
+                    if not skip_max_targets_warning:
+                        l.warning(
+                            "Exit state has over %d possible solutions. Likely unconstrained; skipping. %s",
+                            max_targets,
+                            target.shallow_repr()
+                        )
                     self.unconstrained_successors.append(state)
                 else:
                     for cond, a in cond_and_targets:
@@ -510,7 +517,7 @@ class SimSuccessors(object):
 
 
 from ..state_plugins.inspect import BP_BEFORE, BP_AFTER
-from ..errors import SimSolverModeError, AngrUnsupportedSyscallError, SimValueError
+from ..errors import SimSolverModeError, AngrUnsupportedSyscallError, AngrSyscallError, SimValueError, SimUnsatError
 from ..calling_conventions import SYSCALL_CC
 from ..state_plugins.sim_action_object import _raw_ast
 from ..state_plugins.callstack import CallStack
