@@ -1,21 +1,34 @@
-
 import os
 import logging
 import networkx
 import string
 import itertools
 from collections import defaultdict
+from typing import Union, Optional, Iterable
+from typing import Type # For some reasons the linter doesn't recognize the use in apply_definition but PyCharm needs it imported to correctly recognize it # pylint: disable=unused-import
+
 from itanium_demangler import parse
 
+from cle.backends.symbol import Symbol
 from archinfo.arch_arm import get_real_address_if_arm
 import claripy
-from ...errors import SimEngineError, SimMemoryError
+
+from ...codenode import CodeNode, BlockNode, HookNode, SyscallNode
+from ...serializable import Serializable
+from ...errors import AngrValueError, SimEngineError, SimMemoryError
 from ...procedures import SIM_LIBRARIES
+from ...protos import function_pb2
+from ...calling_conventions import DEFAULT_CC
+from .function_parser import FunctionParser
 
 l = logging.getLogger(name=__name__)
 
+from ...sim_type import SimTypeFunction, parse_defns
+from ...calling_conventions import SimCC
+from ...project import Project
 
-class Function:
+
+class Function(Serializable):
     """
     A representation of a function and various information about it.
     """
@@ -24,19 +37,29 @@ class Function:
                  '_callout_sites', '_endpoints', '_call_sites', '_retout_sites', 'addr', '_function_manager',
                  'is_syscall', '_project', 'is_plt', 'addr', 'is_simprocedure', '_name', 'binary_name',
                  '_argument_registers', '_argument_stack_variables',
-                 'bp_on_stack', 'retaddr_on_stack', 'sp_delta', 'calling_convention', 'prototype', '_returning',
+                 'bp_on_stack', 'retaddr_on_stack', 'sp_delta', '_cc', '_prototype', '_returning',
                  'prepared_registers', 'prepared_stack_variables', 'registers_read_afterwards',
                  'startpoint', '_addr_to_block_node', '_block_sizes', '_block_cache', '_local_blocks',
-                 '_local_block_addrs', 'info', 'tags',
+                 '_local_block_addrs', 'info', 'tags', 'alignment',
                  )
 
-    def __init__(self, function_manager, addr, name=None, syscall=None):
+    def __init__(self, function_manager, addr, name=None, syscall=None, is_simprocedure=None, binary_name=None,
+                 is_plt=None, returning=None, alignment=False):
         """
-        Function constructor
+        Function constructor. If the optional parameters are not provided, they will be automatically determined upon
+        the creation of a Function object.
 
         :param addr:            The address of the function.
-        :param name:            (Optional) The name of the function.
-        :param syscall:         (Optional) Whether this function is a syscall or not.
+
+        The following parameters are optional.
+
+        :param str name:        The name of the function.
+        :param bool syscall:    Whether this function is a syscall or not.
+        :param bool is_simprocedure:    Whether this function is a SimProcedure or not.
+        :param str binary_name: Name of the binary where this function is.
+        :param bool is_plt:     If this function is a PLT entry.
+        :param bool returning:  If this function returns.
+        :param bool alignment:  If this function acts as an alignment filler. Such functions usually only contain nops.
         """
         self.transition_graph = networkx.DiGraph()
         self._local_transition_graph = None
@@ -56,102 +79,124 @@ class Function:
 
         self._call_sites = {}
         self.addr = addr
+        # startpoint can be None if the corresponding CFGNode is a syscall node
+        self.startpoint = None
         self._function_manager = function_manager
-        self.is_syscall = syscall
-
-        self._project = project = self._function_manager._kb._project
-
-        self.is_plt = False
+        self.is_syscall = None
+        self.is_plt = None
         self.is_simprocedure = False
+        self.alignment = alignment
 
-        if self.is_syscall is None:
-            # Determine whether this function is a syscall or not
-            self.is_syscall = self._project.simos.is_syscall_addr(addr)
+        # These properties are set by VariableManager
+        self.bp_on_stack = False
+        self.retaddr_on_stack = False
+        self.sp_delta = 0
+        # Calling convention
+        self._cc = None  # type: Optional[SimCC]
+        # Function prototype
+        self._prototype = None  # type: Optional[SimTypeFunction]
+        # Whether this function returns or not. `None` means it's not determined yet
+        self._returning = None
+        self.prepared_registers = set()
+        self.prepared_stack_variables = set()
+        self.registers_read_afterwards = set()
 
-        # Determine whether this function is a simprocedure
-        if self.is_syscall or project.is_hooked(addr):
-            self.is_simprocedure = True
+        self._addr_to_block_node = {}  # map addresses to nodes. it's a cache of blocks. if a block is removed from the
+                                       # function, it may not be removed from _addr_to_block_node. if you want to list
+                                       # all blocks of a function, access .blocks.
+        self._block_sizes = {}  # map addresses to block sizes
+        self._block_cache = {}  # a cache of real, hard data Block objects
+        self._local_blocks = {}  # a dict of all blocks inside the function
+        self._local_block_addrs = set()  # a set of addresses of all blocks inside the function
 
-        if project.loader.find_plt_stub_name(addr) is not None:
-            # Whether this function is a plt entry or not is fully relying on the PLT detection in CLE
-            self.is_plt = True
-
-        # Try to get a name from existing labels
-        if name is None and addr in function_manager._kb.labels:
-            name = function_manager._kb.labels[addr]
-
-        # try to get the name from a hook
-        if name is None:
-            if project.is_hooked(addr):
-                hooker = project.hooked_by(addr)
-                name = hooker.display_name
-            elif project.simos.is_syscall_addr(addr):
-                syscall_inst = project.simos.syscall_from_addr(addr)
-                name = syscall_inst.display_name
-
-        # generate an IDA-style sub_X name
-        if name is None:
-            name = 'sub_%x' % addr
-
-        binary_name = None
-        # if this function is a simprocedure but not a syscall, use its library name as
-        # its binary name
-        # if it is a syscall, fall back to use self.binary.binary which explicitly says cle##kernel
-        if self.is_simprocedure and not self.is_syscall:
-            hooker = project.hooked_by(addr)
-            if hooker is not None:
-                binary_name = hooker.library_name
-
-        if binary_name is None and self.binary is not None:
-            binary_name = os.path.basename(self.binary.binary)
-
-        self._name = name
-        self.binary_name = binary_name
+        self.info = {}  # storing special information, like $gp values for MIPS32
+        self.tags = tuple()  # store function tags. can be set manually by performing CodeTagging analysis.
 
+        # TODO: Can we remove the following two members?
         # Register offsets of those arguments passed in registers
         self._argument_registers = []
         # Stack offsets of those arguments passed in stack variables
         self._argument_stack_variables = []
 
-        # These properties are set by VariableManager
-        self.bp_on_stack = False
-        self.retaddr_on_stack = False
+        self._project = None  # type: Optional[Project] # will be initialized upon the first access to self.project
 
-        self.sp_delta = 0
+        #
+        # Initialize unspecified properties
+        #
 
-        # Calling convention
-        self.calling_convention = None
+        if syscall is not None:
+            self.is_syscall = syscall
+        else:
+            if self.project is None:
+                raise ValueError("'syscall' must be specified if you do not specify a function manager for this new"
+                                 " function." )
 
-        # Function prototype
-        self.prototype = None
+            # Determine whether this function is a syscall or not
+            self.is_syscall = self.project.simos.is_syscall_addr(addr)
 
-        # Whether this function returns or not. `None` means it's not determined yet
-        self._returning = None
+        # Determine whether this function is a SimProcedure
+        if is_simprocedure is not None:
+            self.is_simprocedure = is_simprocedure
+        else:
+            if self.project is None:
+                raise ValueError("'is_simprocedure' must be specified if you do not specify a function manager for this"
+                                 " new function.")
+
+            if self.is_syscall or self.project.is_hooked(addr):
+                self.is_simprocedure = True
+
+        # Determine if this function is a PLT entry
+        if is_plt is not None:
+            self.is_plt = is_plt
+        else:
+            # Whether this function is a PLT entry or not is fully relying on the PLT detection in CLE
+            if self.project is None:
+                raise ValueError("'is_plt' must be specified if you do not specify a function manager for this new"
+                                 " function.")
+
+            self.is_plt = self.project.loader.find_plt_stub_name(addr) is not None
+
+        # Determine the name of this function
+        if name is None:
+            self._name = self._get_initial_name()
+        else:
+            self._name = name
+
+        # Determine the name the binary where this function is.
+        if binary_name is not None:
+            self.binary_name = binary_name
+        else:
+            self.binary_name = self._get_initial_binary_name()
 
         # Determine returning status for SimProcedures and Syscalls
-        hooker = None
-        if self.is_syscall:
-            hooker = project.simos.syscall_from_addr(addr)
-        elif self.is_simprocedure:
-            hooker = project.hooked_by(addr)
-        if hooker and hasattr(hooker, 'NO_RET'):
-            self.returning = not hooker.NO_RET
+        if returning is not None:
+            self._returning = returning
+        else:
+            if self.project is None:
+                raise ValueError("'returning' must be specified if you do not specify a functio nmnager for this new"
+                                 " function.")
 
-        self.prepared_registers = set()
-        self.prepared_stack_variables = set()
-        self.registers_read_afterwards = set()
+            self._returning = self._get_initial_returning()
 
-        # startpoint can always be None if this CFGNode is a syscall node
-        self.startpoint = None
+        # Determine a calling convention
+        # If it is a SimProcedure it might have a CC already defined which can be used
+        if self.is_simprocedure and self.project is not None and self.addr in self.project._sim_procedures:
+            simproc = self.project._sim_procedures[self.addr]
+            cc = simproc.cc
+            if cc is None:
+                arch = self.project.arch
+                if self.project.arch.name in DEFAULT_CC:
+                    cc = DEFAULT_CC[arch.name](arch)
 
-        self._addr_to_block_node = {}  # map addresses to nodes
-        self._block_sizes = {}  # map addresses to block sizes
-        self._block_cache = {}  # a cache of real, hard data Block objects
-        self._local_blocks = {} # a dict of all blocks inside the function
-        self._local_block_addrs = set()  # a set of addresses of all blocks inside the function
+            # update cc.args according to num_args
+            # TODO: Handle non-traditional arguments like fp
+            if cc is not None and not cc.args and simproc.num_args:
+                args = cc.arg_locs(is_fp=[False] * simproc.num_args)  # arg_locs() uses cc.args
+                cc.args = args
 
-        self.info = { }  # storing special information, like $gp values for MIPS32
-        self.tags = tuple()  # store function tags. can be set manually by performing CodeTagging analysis.
+            self.calling_convention = cc
+        else:
+            self.calling_convention = None
 
     @property
     def name(self):
@@ -162,6 +207,14 @@ class Function:
         self._name = v
         self._function_manager._kb.labels[self.addr] = v
 
+    @property
+    def project(self):
+        if self._project is None:
+            # try to set it from function manager
+            if self._function_manager is not None:
+                self._project = self._function_manager._kb._project #type: Optional[Project]
+        return self._project
+
     @property
     def returning(self):
         return self._returning
@@ -266,6 +319,84 @@ class Function:
         # TODO: remove link register values
         return [const.value for block in self.blocks for const in block.vex.constants]
 
+    @property
+    def calling_convention(self):
+        """
+        Get the calling convention of this function.
+
+        :return:    The calling convention of this function.
+        :rtype:     Optional[SimCC]
+        """
+        return self._cc
+
+    @calling_convention.setter
+    def calling_convention(self, v):
+        """
+        Set the calling convention of this function. If the new cc has a function prototype, we will clear
+        self._prototype. Otherwise, if self.prototype is set, we will use it to update the function prototype of the new
+        cc, and then clear self._prototype. A warning message will be generated in either case.
+
+        :param Optional[SimCC] v:   The new calling convention.
+        :return:                    None
+        """
+        self._cc = v
+
+        if self._cc is not None:
+            if self._cc.func_ty is None and self._prototype is not None:
+                l.warning("The new calling convention for %r does not have a prototype associated. Using the existing "
+                          "function prototype to update the new calling convention. The existing function prototype "
+                          "will be removed.", self)
+                self._cc.func_ty = self._prototype
+                self._prototype = None
+            elif self._cc.func_ty is not None and self._prototype is not None:
+                l.warning("The new calling convention for %r already has a prototype associated. The existing function "
+                          "prototype will be removed.", self)
+                self._prototype = None
+
+    @property
+    def prototype(self):
+        """
+        Get the prototype of this function. We prioritize the function prototype that is set in self.calling_convention.
+
+        :return:    The function prototype.
+        :rtype:     Optional[SimTypeFunction]
+        """
+        if self._cc:
+            return self._cc.func_ty
+        else:
+            return self._prototype
+
+    @prototype.setter
+    def prototype(self, proto):
+        """
+        Set a new prototype to this function. If a calling convention is already set to this function, the new prototype
+        will be set to this calling convention instead.
+
+        :param Optional[SimTypeFunction] proto: The new prototype.
+        :return:    None
+        """
+        if self._cc:
+            self._cc.func_ty = proto.with_arch(self.project.arch) if proto else None
+        else:
+            self._prototype = proto
+
+    @classmethod
+    def _get_cmsg(cls):
+        return function_pb2.Function()
+
+    def serialize_to_cmessage(self):
+        return FunctionParser.serialize(self)
+
+    @classmethod
+    def parse_from_cmessage(cls, cmsg, **kwargs):
+        """
+        :param cmsg:
+
+        :return Function: The function instantiated out of the cmsg data.
+        """
+        return FunctionParser.parse_from_cmsg(cmsg, **kwargs)
+
+
     def string_references(self, minimum_length=2, vex_only=False):
         """
         All of the constant string references used by this function.
@@ -437,6 +568,7 @@ class Function:
         s += '  SP difference: %d\n' % self.sp_delta
         s += '  Has return: %s\n' % self.has_return
         s += '  Returning: %s\n' % ('Unknown' if self.returning is None else self.returning)
+        s += '  Alignment: %s\n' % (self.alignment)
         s += '  Arguments: reg: %s, stack: %s\n' % \
             (self._argument_registers,
              self._argument_stack_variables)
@@ -446,8 +578,9 @@ class Function:
 
     def __repr__(self):
         if self.is_syscall:
-            return '<Syscall function %s (%s)>' % (self.name, self.addr)
-        return '<Function %s (%s)>' % (self.name, self.addr)
+            return '<Syscall function %s (%s)>' % (self.name,
+                                                   hex(self.addr) if isinstance(self.addr, int) else self.addr)
+        return '<Function %s (%s)>' % (self.name, hex(self.addr) if isinstance(self.addr, int) else self.addr)
 
     @property
     def endpoints(self):
@@ -486,6 +619,20 @@ class Function:
 
         return self._project.loader.find_object_containing(self.addr, membership_check=False)
 
+    @property
+    def offset(self) -> int:
+        """
+        :return: the function's binary offset (i.e., non-rebased address)
+        """
+        return self.addr - self.binary.mapped_base
+
+    @property
+    def symbol(self) -> Union[None, Symbol]:
+        """
+        :return: the function's Symbol, if any
+        """
+        return self.binary.loader.find_symbol(self.addr)
+
     def add_jumpout_site(self, node):
         """
         Add a custom jumpout site.
@@ -518,6 +665,79 @@ class Function:
         self._retout_sites.add(node)
         self._add_endpoint(node, 'return')
 
+    def _get_initial_name(self):
+        """
+        Determine the most suitable name of the function.
+
+        :return:    The initial function name.
+        :rtype:     string
+        """
+
+        name = None
+        addr = self.addr
+
+        # Try to get a name from existing labels
+        if self._function_manager is not None:
+            if addr in self._function_manager._kb.labels:
+                name = self._function_manager._kb.labels[addr]
+
+        # try to get the name from a hook
+        if name is None and self.project is not None:
+            project = self.project
+            if project.is_hooked(addr):
+                hooker = project.hooked_by(addr)
+                name = hooker.display_name
+            elif project.simos.is_syscall_addr(addr):
+                syscall_inst = project.simos.syscall_from_addr(addr)
+                name = syscall_inst.display_name
+
+        # generate an IDA-style sub_X name
+        if name is None:
+            name = 'sub_%x' % addr
+
+        return name
+
+    def _get_initial_binary_name(self):
+        """
+        Determine the name of the binary where this function is.
+
+        :return: None
+        """
+
+        binary_name = None
+
+        # if this function is a simprocedure but not a syscall, use its library name as
+        # its binary name
+        # if it is a syscall, fall back to use self.binary.binary which explicitly says cle##kernel
+        if self.project and self.is_simprocedure and not self.is_syscall:
+            hooker = self.project.hooked_by(self.addr)
+            if hooker is not None:
+                binary_name = hooker.library_name
+
+        if binary_name is None and self.binary is not None:
+            binary_name = os.path.basename(self.binary.binary)
+
+        return binary_name
+
+    def _get_initial_returning(self):
+        """
+        Determine if this function returns or not *if it is hooked by a SimProcedure or a user hook*.
+
+        :return:    True if the hooker returns, False otherwise.
+        :rtype:     bool
+        """
+
+        hooker = None
+        if self.is_syscall:
+            hooker = self.project.simos.syscall_from_addr(self.addr)
+        elif self.is_simprocedure:
+            hooker = self.project.hooked_by(self.addr)
+        if hooker and hasattr(hooker, 'NO_RET'):
+            return not hooker.NO_RET
+
+        # Cannot determine
+        return None
+
     def _clear_transition_graph(self):
         self._block_cache = {}
         self._block_sizes = {}
@@ -541,7 +761,7 @@ class Function:
 
         self.transition_graph[src][dst]['confirmed'] = True
 
-    def _transit_to(self, from_node, to_node, outside=False, ins_addr=None, stmt_idx=None):
+    def _transit_to(self, from_node, to_node, outside=False, ins_addr=None, stmt_idx=None, is_exception=False):
         """
         Registers an edge between basic blocks in this function's transition graph.
         Arguments are CodeNode objects.
@@ -566,14 +786,15 @@ class Function:
             else:
                 self._register_nodes(True, from_node)
 
+        type_ = 'transition' if not is_exception else 'exception'
         if to_node is not None:
-            self.transition_graph.add_edge(from_node, to_node, type='transition', outside=outside, ins_addr=ins_addr,
+            self.transition_graph.add_edge(from_node, to_node, type=type_, outside=outside, ins_addr=ins_addr,
                                            stmt_idx=stmt_idx
                                            )
 
         if outside:
             # this node is an endpoint of the current function
-            self._add_endpoint(from_node, 'transition')
+            self._add_endpoint(from_node, type_)
 
         # clear the cache
         self._local_transition_graph = None
@@ -626,7 +847,7 @@ class Function:
         self._local_transition_graph = None
 
     def _return_from_call(self, from_func, to_node, to_outside=False):
-        self.transition_graph.add_edge(from_func, to_node, type='real_return', to_outside=to_outside)
+        self.transition_graph.add_edge(from_func, to_node, type='return', to_outside=to_outside)
         for _, _, data in self.transition_graph.in_edges(to_node, data=True):
             if 'type' in data and data['type'] == 'fake_return':
                 data['confirmed'] = True
@@ -639,6 +860,8 @@ class Function:
 
         for node in nodes:
             self.transition_graph.add_node(node)
+            if not isinstance(node, CodeNode):
+                continue
             node._graph = self.transition_graph
             if node.addr not in self or self._block_sizes[node.addr] == 0:
                 self._block_sizes[node.addr] = node.size
@@ -752,11 +975,11 @@ class Function:
                         self._callout_sites.add(the_node)
                         self._add_endpoint(the_node, 'call')
 
-    def get_call_sites(self):
+    def get_call_sites(self) -> Iterable[int]:
         """
         Gets a list of all the basic blocks that end in calls.
 
-        :return:                    A list of the addresses of the blocks that end in calls.
+        :return:                    A view of the addresses of the blocks that end in calls.
         """
         return self._call_sites.keys()
 
@@ -787,7 +1010,14 @@ class Function:
     @property
     def graph(self):
         """
-        Return a local transition graph that only contain nodes in current function.
+        Get a local transition graph. A local transition graph is a transition graph that only contains nodes that
+        belong to the current function. All edges, except for the edges going out from the current function or coming
+        from outside the current function, are included.
+
+        The generated graph is cached in self._local_transition_graph.
+
+        :return:    A local transition graph.
+        :rtype:     networkx.DiGraph
         """
 
         if self._local_transition_graph is not None:
@@ -800,7 +1030,7 @@ class Function:
             g.add_node(block)
         for src, dst, data in self.transition_graph.edges(data=True):
             if 'type' in data:
-                if data['type']  == 'transition' and ('outside' not in data or data['outside'] is False):
+                if data['type'] in ('transition', 'exception') and ('outside' not in data or data['outside'] is False):
                     g.add_edge(src, dst, **data)
                 elif data['type'] == 'fake_return' and 'confirmed' in data and \
                         ('outside' not in data or data['outside'] is False):
@@ -810,13 +1040,99 @@ class Function:
 
         return g
 
+    def graph_ex(self, exception_edges=True):
+        """
+        Get a local transition graph with a custom configuration. A local transition graph is a transition graph that
+        only contains nodes that belong to the current function. This method allows user to exclude certain types of
+        edges together with the nodes that are only reachable through such edges, such as exception edges.
+
+        The generated graph is not cached.
+
+        :param bool exception_edges:    Should exception edges and the nodes that are only reachable through exception
+                                        edges be kept.
+        :return:                        A local transition graph with a special configuration.
+        :rtype:                         networkx.DiGraph
+        """
+
+        # graph_ex() should not impact any already cached graph
+        old_cached_graph = self._local_transition_graph
+        graph = self.graph
+        self._local_transition_graph = old_cached_graph  # restore the cached graph
+
+        # fast path
+        if exception_edges:
+            return graph
+
+        # BFS on local graph but ignoring certain types of graphs
+        g = networkx.DiGraph()
+        queue = [ n for n in graph if n is self.startpoint or graph.in_degree[n] == 0 ]
+        traversed = set(queue)
+
+        while queue:
+            node = queue.pop(0)
+
+            g.add_node(node)
+            for _, dst, edge_data in graph.out_edges(node, data=True):
+                edge_type = edge_data.get('type', None)
+                if not exception_edges and edge_type == 'exception':
+                    # ignore this edge
+                    continue
+                g.add_edge(node, dst, **edge_data)
+
+                if dst not in traversed:
+                    traversed.add(dst)
+                    queue.append(dst)
+
+        return g
+
+    def transition_graph_ex(self, exception_edges=True):
+        """
+        Get a transition graph with a custom configuration. This method allows user to exclude certain types of edges
+        together with the nodes that are only reachable through such edges, such as exception edges.
+
+        The generated graph is not cached.
+
+        :param bool exception_edges:    Should exception edges and the nodes that are only reachable through exception
+                                        edges be kept.
+        :return:                        A local transition graph with a special configuration.
+        :rtype:                         networkx.DiGraph
+        """
+
+        graph = self.transition_graph
+
+        # fast path
+        if exception_edges:
+            return graph
+
+        # BFS on local graph but ignoring certain types of graphs
+        g = networkx.DiGraph()
+        queue = [ n for n in graph if n is self.startpoint or graph.in_degree[n] == 0 ]
+        traversed = set(queue)
+
+        while queue:
+            node = queue.pop(0)
+            traversed.add(node)
+
+            g.add_node(node)
+            for _, dst, edge_data in graph.out_edges(node, data=True):
+                edge_type = edge_data.get('type', None)
+                if not exception_edges and edge_type == 'exception':
+                    # ignore this edge
+                    continue
+                g.add_edge(node, dst, **edge_data)
+
+                if dst not in traversed:
+                    traversed.add(dst)
+                    queue.append(dst)
+
+        return g
+
     def subgraph(self, ins_addrs):
         """
         Generate a sub control flow graph of instruction addresses based on self.graph
 
         :param iterable ins_addrs: A collection of instruction addresses that should be included in the subgraph.
-        :return: A subgraph.
-        :rtype: networkx.DiGraph
+        :return networkx.DiGraph: A subgraph.
         """
 
         # find all basic blocks that include those instructions
@@ -859,8 +1175,7 @@ class Function:
         Get the size of the instruction specified by `insn_addr`.
 
         :param int insn_addr: Address of the instruction
-        :return: Size of the instruction in bytes, or None if the instruction is not found.
-        :rtype: int
+        :return int: Size of the instruction in bytes, or None if the instruction is not found.
         """
 
         for b in self.blocks:
@@ -876,6 +1191,29 @@ class Function:
 
         return None
 
+    def addr_to_instruction_addr(self, addr):
+        """
+        Obtain the address of the instruction that covers @addr.
+
+        :param int addr:    An address.
+        :return:            Address of the instruction that covers @addr, or None if this addr is not covered by any
+                            instruction of this function.
+        :rtype:             int or None
+        """
+
+        # TODO: Replace the linear search with binary search
+        for b in self.blocks:
+            if b.addr <= addr < b.addr + b.size:
+                # found it
+                for i, instr_addr in enumerate(b.instruction_addrs):
+                    if i < len(b.instruction_addrs) - 1 and instr_addr <= addr < b.instruction_addrs[i+1]:
+                        return instr_addr
+                    elif i == len(b.instruction_addrs) - 1 and instr_addr <= addr:
+                        return instr_addr
+                # Not covered by any instruction... why?
+                return None
+        return None
+
     def dbg_print(self):
         """
         Returns a representation of the list of basic blocks in this function.
@@ -1073,6 +1411,7 @@ class Function:
             # PLT entries must have the same declaration as their jump targets
             # Try to determine which library this PLT entry will jump to
             edges = self.transition_graph.edges()
+            if len(edges) == 0: return
             node = next(iter(edges))[1]
             if len(edges) == 1 and (type(node) is HookNode or type(node) is SyscallNode):
                 target = node.addr
@@ -1098,7 +1437,8 @@ class Function:
             self.calling_convention.args = None
             self.calling_convention.func_ty = proto
 
-    def _addr_to_funcloc(self, addr):
+    @staticmethod
+    def _addr_to_funcloc(addr):
 
         # FIXME
         if isinstance(addr, tuple):
@@ -1106,7 +1446,6 @@ class Function:
         else:  # int, long
             return addr
 
-
     @property
     def demangled_name(self):
 
@@ -1116,6 +1455,37 @@ class Function:
                 return ast.__str__()
         return self.name
 
+    def apply_definition(self, definition, calling_convention=None):
+        """
+
+        :param str definition:
+        :param Optional[Union[SimCC, Type[SimCC]]] calling_convention:
+        :return None:
+        """
+        if not definition.endswith(";"):
+            definition += ";"
+        func_def = parse_defns(definition)
+        if len(func_def.keys()) > 1:
+            raise Exception("Too many definitions: %s " % list(func_def.keys()))
+
+        name, ty = func_def.popitem() # type: str, SimTypeFunction
+        self.name = name
+        # setup the calling convention
+        # If a SimCC object is passed assume that this is sane and just use it
+        if isinstance(calling_convention, SimCC):
+            self.calling_convention = calling_convention
+
+        # If it is a subclass of SimCC we can instantiate it
+        elif isinstance(calling_convention, type) and issubclass(calling_convention, SimCC):
+            self.calling_convention = calling_convention(self.project.arch, func_ty=ty)
+
+        # If none is specified default to something
+        elif calling_convention is None:
+            self.calling_convention = self.project.factory.cc(func_ty=ty)
+
+        else:
+            raise TypeError("calling_convention has to be one of: [SimCC, type(SimCC), None]")
+
     def copy(self):
         func = Function(self._function_manager, self.addr, name=self.name, syscall=self.is_syscall)
         func.transition_graph = networkx.DiGraph(self.transition_graph)
@@ -1135,6 +1505,7 @@ class Function:
         func.calling_convention = self.calling_convention
         func.prototype = self.prototype
         func._returning = self._returning
+        func.alignment = self.alignment
         func.startpoint = self.startpoint
         func._addr_to_block_node = self._addr_to_block_node.copy()
         func._block_sizes = self._block_sizes.copy()
@@ -1145,7 +1516,3 @@ class Function:
         func.tags = self.tags
 
         return func
-
-
-from ...codenode import BlockNode, HookNode, SyscallNode
-from ...errors import AngrValueError
