@@ -3,6 +3,9 @@ import os
 import archinfo
 from collections import defaultdict
 import logging
+import inspect
+
+import itanium_demangler
 
 from ...calling_conventions import DEFAULT_CC
 from ...misc import autoimport
@@ -13,7 +16,8 @@ from ..stubs.syscall_stub import syscall as stub_syscall
 l = logging.getLogger(name=__name__)
 SIM_LIBRARIES = {}
 
-class SimLibrary(object):
+
+class SimLibrary:
     """
     A SimLibrary is the mechanism for describing a dynamic library's API, its functions and metadata.
 
@@ -168,10 +172,15 @@ class SimLibrary(object):
     def _apply_metadata(self, proc, arch):
         if proc.cc is None and arch.name in self.default_ccs:
             proc.cc = self.default_ccs[arch.name](arch)
+            if proc.cc.func_ty is not None:
+                # Use inspect to extract the parameters from the run python function
+                proc.cc.func_ty.arg_names = inspect.getfullargspec(proc.run).args[1:]
         if proc.display_name in self.prototypes:
             if proc.cc is None:
                 proc.cc = self.fallback_cc[arch.name](arch)
             proc.cc.func_ty = self.prototypes[proc.display_name].with_arch(arch)
+            # Use inspect to extract the parameters from the run python function
+            proc.cc.func_ty.arg_names = inspect.getfullargspec(proc.run).args[1:]
             if not proc.ARGS_MISMATCH:
                 proc.cc.num_args = len(proc.cc.func_ty.args)
                 proc.num_args = len(proc.cc.func_ty.args)
@@ -244,6 +253,76 @@ class SimLibrary(object):
         return func_name in self.prototypes
 
 
+class SimCppLibrary(SimLibrary):
+    """
+    SimCppLibrary is a specialized version of SimLibrary that will demangle C++ function names before looking for an
+    implementation or prototype for it.
+    """
+
+    @staticmethod
+    def _try_demangle(name):
+        if name[0:2] == "_Z":
+            ast = itanium_demangler.parse(name)
+            if ast:
+                return str(ast)
+        return name
+
+    def get(self, name, arch):
+        """
+        Get an implementation of the given function specialized for the given arch, or a stub procedure if none exists.
+        Demangle the function name if it is a mangled C++ name.
+
+        :param str name:    The name of the function as a string
+        :param arch:    The architecure to use, as either a string or an archinfo.Arch instance
+        :return:        A SimProcedure instance representing the function as found in the library
+        """
+        name = self._try_demangle(name)
+        return super().get(name, arch)
+
+    def get_stub(self, name, arch):
+        """
+        Get a stub procedure for the given function, regardless of if a real implementation is available. This will
+        apply any metadata, such as a default calling convention or a function prototype. Demangle the function name
+        if it is a mangled C++ name.
+
+        :param str name:    The name of the function as a string
+        :param arch:        The architecture to use, as either a string or an archinfo.Arch instance
+        :return:            A SimProcedure instance representing a plausable stub as could be found in the library.
+        """
+        name = self._try_demangle(name)
+        return super().get_stub(name, arch)
+
+    def has_metadata(self, name):
+        """
+        Check if a function has either an implementation or any metadata associated with it. Demangle the function name
+        if it is a mangled C++ name.
+
+        :param name:    The name of the function as a string
+        :return:        A bool indicating if anything is known about the function
+        """
+        name = self._try_demangle(name)
+        return super().has_metadata(name)
+
+    def has_implementation(self, name):
+        """
+        Check if a function has an implementation associated with it. Demangle the function name if it is a mangled C++
+        name.
+
+        :param str name:    A mangled function name.
+        :return:            bool
+        """
+        return super().has_implementation(self._try_demangle(name))
+
+    def has_prototype(self, func_name):
+        """
+        Check if a function has a prototype associated with it. Demangle the function name if it is a mangled C++ name.
+
+        :param str name:    A mangled function name.
+        :return:            bool
+        """
+        return super().has_prototype(self._try_demangle(func_name))
+
+
 class SimSyscallLibrary(SimLibrary):
     """
     SimSyscallLibrary is a specialized version of SimLibrary for dealing not with a dynamic library's API but rather
