@@ -1,12 +1,18 @@
 from collections import OrderedDict, defaultdict
+from .misc.ux import deprecated
 import copy
 import re
 import logging
+from typing import Optional
 
 import claripy
 
 l = logging.getLogger(name=__name__)
 
+# pycparser hack to parse type expressions
+errorlog = logging.getLogger(name=__name__ + ".yacc")
+errorlog.setLevel(logging.ERROR)
+
 try:
     import pycparser
 except ImportError:
@@ -79,6 +85,8 @@ class SimType:
         """
         if self._arch is None:
             return NotImplemented
+        if self.size is NotImplemented:
+            return NotImplemented
         return self.size // self._arch.byte_width
 
     def with_arch(self, arch):
@@ -95,17 +103,28 @@ class SimType:
     def _init_str(self):
         return "NotImplemented(%s)" % (self.__class__.__name__)
 
+    def c_repr(self):
+        raise NotImplementedError()
+
 
 class SimTypeBottom(SimType):
     """
     SimTypeBottom basically represents a type error.
     """
 
-    def __repr__(self):
+    def __repr__(self, label=None):
         return 'BOT'
 
     def _init_str(self):
-        return "%s()" % self.__class__.__name__
+        return "%s(%s)" % (
+            self.__class__.__name__,
+            ("label=\"%s\"" % self.label) if self.label else ""
+        )
+
+    def c_repr(self):
+        if self.label:
+            return self.label
+        return "BOT"
 
 
 class SimTypeTop(SimType):
@@ -122,6 +141,9 @@ class SimTypeTop(SimType):
     def __repr__(self):
         return 'TOP'
 
+    def c_repr(self):
+        return "TOP"
+
 
 class SimTypeReg(SimType):
     """
@@ -153,7 +175,7 @@ class SimTypeReg(SimType):
     def store(self, state, addr, value):
         store_endness = state.arch.memory_endness
 
-        if isinstance(value, claripy.ast.Bits):
+        if isinstance(value, claripy.ast.Bits):  # pylint:disable=isinstance-second-argument-not-valid-type
             if value.size() != self.size:
                 raise ValueError("size of expression is wrong size for type")
         elif isinstance(value, int):
@@ -165,6 +187,9 @@ class SimTypeReg(SimType):
 
         state.memory.store(addr, value, endness=store_endness)
 
+    def c_repr(self):
+        return "<Reg_%d>" % self.size
+
 
 class SimTypeNum(SimType):
     """
@@ -183,6 +208,9 @@ class SimTypeNum(SimType):
         self._size = size
         self.signed = signed
 
+    def c_repr(self):
+        return "{}int{}_t".format('' if self.signed else 'u', self.size)
+
     def __repr__(self):
         return "{}int{}_t".format('' if self.signed else 'u', self.size)
 
@@ -198,7 +226,7 @@ class SimTypeNum(SimType):
     def store(self, state, addr, value):
         store_endness = state.arch.memory_endness
 
-        if isinstance(value, claripy.ast.Bits):
+        if isinstance(value, claripy.ast.Bits):  # pylint:disable=isinstance-second-argument-not-valid-type
             if value.size() != self.size:
                 raise ValueError("size of expression is wrong size for type")
         elif isinstance(value, int):
@@ -227,6 +255,12 @@ class SimTypeInt(SimTypeReg):
         super(SimTypeInt, self).__init__(None, label=label)
         self.signed = signed
 
+    def c_repr(self):
+        name = self._base_name
+        if not self.signed:
+            name = 'unsigned ' + name
+        return name
+
     def __repr__(self):
         name = self._base_name
         if not self.signed:
@@ -256,10 +290,10 @@ class SimTypeInt(SimTypeReg):
         return n
 
     def _init_str(self):
-        return "%s(signed=%s, label=%s)" % (
+        return "%s(signed=%s%s)" % (
             self.__class__.__name__,
             self.signed,
-            '"%s"' % self.label if self.label is not None else "None",
+            (', label="%s"' % self.label) if self.label is not None else "",
         )
 
     def _refine_dir(self):
@@ -295,17 +329,20 @@ class SimTypeChar(SimTypeReg):
     this could be represented by a byte, but this is meant to be interpreted as a character.
     """
 
-    def __init__(self, label=None):
+    def __init__(self, signed=True, label=None):
         """
         :param label: the type label.
         """
         # FIXME: Now the size of a char is state-dependent.
         SimTypeReg.__init__(self, 8, label=label)
-        self.signed = False
+        self.signed = signed
 
     def __repr__(self):
         return 'char'
 
+    def c_repr(self):
+        return "char"
+
     def store(self, state, addr, value):
         # FIXME: This is a hack.
         self._size = state.arch.byte_width
@@ -328,9 +365,9 @@ class SimTypeChar(SimTypeReg):
         return out
 
     def _init_str(self):
-        return "%s(label=%s)" % (
+        return "%s(%s)" % (
             self.__class__.__name__,
-            '"%s"' if self.label is not None else "None",
+            ('label="%s"' % self.label) if self.label is not None else "",
         )
 
 
@@ -350,6 +387,9 @@ class SimTypeBool(SimTypeChar):
     def _init_str(self):
         return "%s()" % (self.__class__.__name__)
 
+    def c_repr(self):
+        return "bool"
+
 
 class SimTypeFd(SimTypeReg):
     """
@@ -369,6 +409,9 @@ class SimTypeFd(SimTypeReg):
     def __repr__(self):
         return 'fd_t'
 
+    def c_repr(self):
+        return "fd_t"
+
 
 class SimTypePointer(SimTypeReg):
     """
@@ -390,6 +433,9 @@ class SimTypePointer(SimTypeReg):
     def __repr__(self):
         return '{}*'.format(self.pts_to)
 
+    def c_repr(self):
+        return '{}*'.format(self.pts_to.c_repr())
+
     def make(self, pts_to):
         new = type(self)(pts_to)
         new._arch = self._arch
@@ -407,10 +453,10 @@ class SimTypePointer(SimTypeReg):
         return out
 
     def _init_str(self):
-        return "%s(%s, label=%s, offset=%d)" % (
+        return "%s(%s%s, offset=%d)" % (
             self.__class__.__name__,
             self.pts_to._init_str(),
-            '"%s"' % self.label if self.label is not None else "None",
+            (', label="%s"' % self.label) if self.label is not None else "",
             self.offset
         )
 
@@ -428,6 +474,9 @@ class SimTypeFixedSizeArray(SimType):
     def __repr__(self):
         return '{}[{}]'.format(self.elem_type, self.length)
 
+    def c_repr(self):
+        return '{}[{}]'.format(self.elem_type, self.length)
+
     _can_refine_int = True
 
     def _refine(self, view, k):
@@ -481,6 +530,9 @@ class SimTypeArray(SimType):
     def __repr__(self):
         return '{}[{}]'.format(self.elem_type, '' if self.length is None else self.length)
 
+    def c_repr(self):
+        return '{}[{}]'.format(self.elem_type, '' if self.length is None else self.length)
+
     @property
     def size(self):
         if self._arch is None:
@@ -515,6 +567,9 @@ class SimTypeString(SimTypeArray):
     def __repr__(self):
         return 'string_t'
 
+    def c_repr(self):
+        return 'string_t'
+
     def extract(self, state, addr, concrete=False):
         if self.length is None:
             out = None
@@ -523,7 +578,7 @@ class SimTypeString(SimTypeArray):
             if state.solver.symbolic(last_byte):
                 raise ValueError("Trying to extract a symbolic string at %#x" % state.solver.eval(addr))
             addr += 1
-            while not claripy.is_true(last_byte == 0):
+            while not (claripy.is_true(last_byte == 0) or state.solver.symbolic(last_byte)):
                 out = last_byte if out is None else out.concat(last_byte)
                 last_byte = state.memory.load(addr, 1)
                 addr += 1
@@ -566,6 +621,9 @@ class SimTypeWString(SimTypeArray):
     def __repr__(self):
         return 'wstring_t'
 
+    def c_repr(self):
+        return 'wstring_t'
+
     def extract(self, state, addr, concrete=False):
         if self.length is None:
             out = None
@@ -574,7 +632,7 @@ class SimTypeWString(SimTypeArray):
             if state.solver.symbolic(last_byte):
                 raise ValueError("Trying to extract a symbolic string at %#x" % state.solver.eval(addr))
             addr += 2
-            while not claripy.is_true(last_byte == 0):
+            while not (claripy.is_true(last_byte == 0) or state.solver.symbolic(last_byte)):
                 out = last_byte if out is None else out.concat(last_byte)
                 last_byte = state.memory.load(addr, 2)
                 addr += 2
@@ -614,17 +672,26 @@ class SimTypeFunction(SimType):
     _fields = ('args', 'returnty')
     base = False
 
-    def __init__(self, args, returnty, label=None):
+    def __init__(self, args, returnty, label=None, arg_names=None, variadic=False):
         """
         :param label:    The type label
         :param args:     A tuple of types representing the arguments to the function
         :param returnty: The return type of the function, or none for void
+        :param variadic: Whether the function accepts varargs
         """
         super(SimTypeFunction, self).__init__(label=label)
         self.args = args
-        self.returnty = returnty
+        self.returnty: Optional[SimType] = returnty
+        self.arg_names = arg_names if arg_names else []
+        self.variadic = variadic
 
     def __repr__(self):
+        argstrs = [str(a) for a in self.args]
+        if self.variadic:
+            argstrs.append('...')
+        return '({}) -> {}'.format(', '.join(argstrs), self.returnty)
+
+    def c_repr(self):
         return '({}) -> {}'.format(', '.join(str(a) for a in self.args), self.returnty)
 
     @property
@@ -632,16 +699,28 @@ class SimTypeFunction(SimType):
         return 4096     # ???????????
 
     def _with_arch(self, arch):
-        out = SimTypeFunction([a.with_arch(arch) for a in self.args], self.returnty.with_arch(arch), self.label)
+        out = SimTypeFunction([a.with_arch(arch) for a in self.args], self.returnty.with_arch(arch),
+                              label=self.label,
+                              arg_names=self.arg_names,
+                              variadic=self.variadic
+                              )
         out._arch = arch
         return out
 
+    def _arg_names_str(self, show_variadic=True):
+        argnames = list(self.arg_names)
+        if self.variadic and show_variadic:
+            argnames.append('...')
+        return ", ".join('"%s"' % arg_name for arg_name in argnames)
+
     def _init_str(self):
-        return "%s([%s], %s, label=%s)" % (
+        return "%s([%s], %s%s%s%s)" % (
             self.__class__.__name__,
             ", ".join([arg._init_str() for arg in self.args]),
             self.returnty._init_str(),
-            self.label
+            (", label=%s" % self.label) if self.label else "",
+            (", arg_names=[%s]" % self._arg_names_str(show_variadic=False)) if self.arg_names else "",
+            ", variadic=True" if self.variadic else "",
         )
 
 
@@ -668,6 +747,9 @@ class SimTypeLength(SimTypeLong):
     def __repr__(self):
         return 'size_t'
 
+    def c_repr(self):
+        return 'size_t'
+
     @property
     def size(self):
         if self._arch is None:
@@ -705,6 +787,15 @@ class SimTypeFloat(SimTypeReg):
     def __repr__(self):
         return 'float'
 
+    def _init_str(self):
+        return "%s(size=%d)" % (
+            self.__class__.__name__,
+            self.size
+        )
+
+    def c_repr(self):
+        return 'float'
+
 
 class SimTypeDouble(SimTypeFloat):
     """
@@ -719,6 +810,9 @@ class SimTypeDouble(SimTypeFloat):
     def __repr__(self):
         return 'double'
 
+    def c_repr(self):
+        return 'double'
+
     @property
     def alignment(self):
         return 8 if self.align_double else 4
@@ -752,6 +846,10 @@ class SimStruct(SimType):
         offsets = {}
         offset_so_far = 0
         for name, ty in self.fields.items():
+            if isinstance(ty, SimTypeBottom):
+                l.warning("Found a bottom field in struct %s. Ignore and increment the offset using the default "
+                          "element size.", self.name)
+                continue
             if not self._pack:
                 align = ty.alignment
                 if offset_so_far % align != 0:
@@ -786,6 +884,12 @@ class SimStruct(SimType):
     def __repr__(self):
         return 'struct %s' % self.name
 
+    def c_repr(self):
+        return 'struct %s' % self.name
+
+    def __hash__(self):
+        return hash((SimStruct, self._name, self._align, self._pack, tuple(self.fields.keys())))
+
     @property
     def size(self):
         return sum(val.size for val in self.fields.values())
@@ -819,11 +923,14 @@ class SimStruct(SimType):
             ty = self.fields[field]
             ty.store(state, addr + offset, value[field])
 
+    @staticmethod
+    def _field_str(field_name, field_type):
+        return "\"%s\": %s" % (field_name, field_type._init_str())
 
     def _init_str(self):
-        return "%s([%s], name=\"%s\", pack=%s, align=%s)" % (
+        return "%s({%s}, name=\"%s\", pack=%s, align=%s)" % (
             self.__class__.__name__,
-            ", ".join([f._init_str() for f in self.fields]),
+            ", ".join([self._field_str(f, ty) for f, ty in self.fields.items()]),
             self._name,
             self._pack,
             self._align,
@@ -856,16 +963,21 @@ class SimStructValue:
 
 
 class SimUnion(SimType):
-    """
-    why
-    """
-    def __init__(self, members, label=None):
+    _fields = ('members', 'name')
+
+    def __init__(self, members, name=None, label=None):
         """
-        :param members:     The members of the struct, as a mapping name -> type
+        :param members:     The members of the union, as a mapping name -> type
+        :param name:        The name of the union
         """
         super(SimUnion, self).__init__(label)
+        self._name = name if name is not None else '<anon>'
         self.members = members
 
+    @property
+    def name(self):
+        return self._name
+
     @property
     def size(self):
         return max(ty.size for ty in self.members.values())
@@ -875,7 +987,15 @@ class SimUnion(SimType):
         return max(val.alignment for val in self.members.values())
 
     def __repr__(self):
-        return 'union {\n\t%s\n}' % '\n\t'.join('%s %s;' % (name, repr(ty)) for name, ty in self.members.items())
+        # use the str instead of repr of each member to avoid exceed recursion
+        # depth when representing self-referential unions
+        return 'union %s {\n\t%s\n}' % (self.name, '\n\t'.join('%s %s;' % (name, str(ty)) for name, ty in self.members.items()))
+
+    def c_repr(self):
+        return 'union %s {\n\t%s\n}' % (self.name, '\n\t'.join('%s %s;' % (name, str(ty)) for name, ty in self.members.items()))
+
+    def __str__(self):
+        return 'union %s' % (self.name, )
 
     def _with_arch(self, arch):
         out = SimUnion({name: ty.with_arch(arch) for name, ty in self.members.items()}, self.label)
@@ -886,7 +1006,7 @@ class SimUnion(SimType):
 BASIC_TYPES = {
     'char': SimTypeChar(),
     'signed char': SimTypeChar(),
-    'unsigned char': SimTypeChar(),
+    'unsigned char': SimTypeChar(signed=False),
 
     'short': SimTypeShort(True),
     'signed short': SimTypeShort(True),
@@ -915,7 +1035,8 @@ BASIC_TYPES = {
 
     'float': SimTypeFloat(),
     'double': SimTypeDouble(),
-    'void': SimTypeBottom(),
+    'long double': SimTypeDouble(),
+    'void': SimTypeBottom(label="void"),
 }
 
 ALL_TYPES = {
@@ -943,6 +1064,8 @@ ALL_TYPES = {
 
     'string': SimTypeString(),
     'wstring': SimTypeWString(),
+
+    'va_list': SimStruct({}, name='va_list')
 }
 
 
@@ -951,7 +1074,27 @@ ALL_TYPES.update(BASIC_TYPES)
 
 # this is a hack, pending https://github.com/eliben/pycparser/issues/187
 def make_preamble():
-    out = []
+    out = ['typedef int TOP;',
+           'typedef struct FILE_t FILE;',
+           'typedef int pid_t;',
+           'typedef int sigset_t;',
+           'typedef int intmax_t;',
+           'typedef unsigned int uintmax_t;',
+           'typedef unsigned int uid_t;',
+           'typedef unsigned int gid_t;',
+           'typedef unsigned int sem_t;',
+           'typedef unsigned short wchar_t;',
+           'typedef unsigned short wctrans_t;',
+           'typedef unsigned short wctype_t;',
+           'typedef unsigned int wint_t;',
+           'typedef unsigned int pthread_key_t;',
+           'typedef long clock_t;',
+           'typedef unsigned int speed_t;',
+           'typedef int socklen_t;',
+           'typedef unsigned short mode_t;',
+           'typedef unsigned long off_t;',
+           'typedef struct va_list {} va_list;',
+           ]
     types_out = []
     for ty in ALL_TYPES:
         if ty in BASIC_TYPES:
@@ -981,7 +1124,26 @@ def make_preamble():
 
     return '\n'.join(out) + '\n', types_out
 
+def _make_scope():
+    """
+    Generate CParser scope_stack argument to parse method
+    """
+    scope = dict()
+    for ty in ALL_TYPES:
+        if ty in BASIC_TYPES:
+            continue
+        if ' ' in ty:
+            continue
 
+        typ = ALL_TYPES[ty]
+        if isinstance(typ, (SimTypeFunction,SimTypeString, SimTypeWString)):
+            continue
+
+        scope[ty] = True
+    return [scope]
+
+
+@deprecated(replacement="register_types(parse_type(struct_expr))")
 def define_struct(defn):
     """
     Register a struct definition globally
@@ -990,23 +1152,37 @@ def define_struct(defn):
     """
     struct = parse_type(defn)
     ALL_TYPES[struct.name] = struct
+    ALL_TYPES['struct ' + struct.name] = struct
     return struct
 
 
-def register_types(mapping):
+def register_types(types):
     """
-    Pass in a mapping from name to SimType and they will be registered to the global type store
+    Pass in some types and they will be registered to the global type store.
+
+    The argument may be either a mapping from name to SimType, or a plain SimType.
+    The plain SimType must be either a struct or union type with a name present.
 
     >>> register_types(parse_types("typedef int x; typedef float y;"))
+    >>> register_types(parse_type("struct abcd { int ab; float cd; }"))
     """
-    ALL_TYPES.update(mapping)
+    if type(types) is SimStruct:
+        if types.name == '<anon>':
+            raise ValueError("Cannot register anonymous struct")
+        ALL_TYPES['struct ' + types.name] = types
+    elif type(types) is SimUnion:
+        if types.name == '<anon>':
+            raise ValueError("Cannot register anonymous union")
+        ALL_TYPES['union ' + types.name] = types
+    else:
+        ALL_TYPES.update(types)
 
 
 def do_preprocess(defn):
     """
     Run a string through the C preprocessor that ships with pycparser but is weirdly inaccessible?
     """
-    from pycparser.ply import lex, cpp
+    from pycparser.ply import lex, cpp  # pylint:disable=import-outside-toplevel
     lexer = lex.lex(cpp)
     p = cpp.Preprocessor(lexer)
     # p.add_path(dir) will add dir to the include search path
@@ -1056,14 +1232,15 @@ def parse_file(defn, preprocess=True):
             if piece.name is not None:
                 out[piece.name] = ty
         elif isinstance(piece, pycparser.c_ast.Typedef):
-            extra_types[piece.name] = _decl_to_type(piece.type, extra_types)
+            extra_types[piece.name] = copy.copy(_decl_to_type(piece.type, extra_types))
+            extra_types[piece.name].label = piece.name
 
     for ty in ignoreme:
         del extra_types[ty]
     return out, extra_types
 
 
-def parse_type(defn, preprocess=True):
+def parse_type(defn, preprocess=True):  # pylint:disable=unused-argument
     """
     Parse a simple type expression into a SimType
 
@@ -1072,27 +1249,61 @@ def parse_type(defn, preprocess=True):
     if pycparser is None:
         raise ImportError("Please install pycparser in order to parse C definitions")
 
-    defn = 'typedef ' + defn.strip('; \n\t\r') + ' QQQQ;'
+    defn = re.sub(r"/\*.*?\*/", r"", defn)
 
-    if preprocess:
-        defn = do_preprocess(defn)
+    parser = pycparser.CParser()
 
-    node = pycparser.c_parser.CParser().parse(make_preamble()[0] + defn)
-    if not isinstance(node, pycparser.c_ast.FileAST) or \
-            not isinstance(node.ext[-1], pycparser.c_ast.Typedef):
+    parser.cparser = pycparser.ply.yacc.yacc(module=parser,
+                                             start='parameter_declaration',
+                                             debug=False,
+                                             optimize=False,
+                                             errorlog=errorlog)
+
+    node = parser.parse(text=defn, scope_stack=_make_scope())
+    if not isinstance(node, pycparser.c_ast.Typename) and \
+            not isinstance(node, pycparser.c_ast.Decl):
         raise ValueError("Something went horribly wrong using pycparser")
 
-    decl = node.ext[-1].type
+    decl = node.type
     return _decl_to_type(decl)
 
+
+def _accepts_scope_stack():
+    """
+    pycparser hack to include scope_stack as parameter in CParser parse method
+    """
+    def parse(self, text, scope_stack=None, filename='', debuglevel=0):
+        self.clex.filename = filename
+        self.clex.reset_lineno()
+        self._scope_stack = [dict()] if scope_stack is None else scope_stack
+        self._last_yielded_token = None
+        return self.cparser.parse(
+            input=text,
+            lexer=self.clex,
+            debug=debuglevel)
+    setattr(pycparser.CParser, 'parse', parse)
+
+
 def _decl_to_type(decl, extra_types=None):
     if extra_types is None: extra_types = {}
 
     if isinstance(decl, pycparser.c_ast.FuncDecl):
-        argtyps = () if decl.args is None else [_decl_to_type(x.type, extra_types) for x in decl.args.params]
-        return SimTypeFunction(argtyps, _decl_to_type(decl.type, extra_types))
+        argtyps = () if decl.args is None else [_decl_to_type(x.type, extra_types) if type(x) is not pycparser.c_ast.EllipsisParam else ... for x in decl.args.params]
+        arg_names = [ arg.name for arg in decl.args.params if type(arg) is not pycparser.c_ast.EllipsisParam] if decl.args else None
+        # special handling: func(void) is func()
+        if len(argtyps) == 1 and isinstance(argtyps[0], SimTypeBottom):
+            argtyps = ()
+            arg_names = None
+        if argtyps and argtyps[-1] is ...:
+            argtyps.pop()
+            variadic = True
+        else:
+            variadic = False
+        return SimTypeFunction(argtyps, _decl_to_type(decl.type, extra_types), arg_names=arg_names, variadic=variadic)
 
     elif isinstance(decl, pycparser.c_ast.TypeDecl):
+        if decl.declname == 'TOP':
+            return SimTypeTop()
         return _decl_to_type(decl.type, extra_types)
 
     elif isinstance(decl, pycparser.c_ast.PtrDecl):
@@ -1109,22 +1320,55 @@ def _decl_to_type(decl, extra_types=None):
         return SimTypeFixedSizeArray(elem_type, size)
 
     elif isinstance(decl, pycparser.c_ast.Struct):
-        struct = SimStruct(OrderedDict(), decl.name)
+        if decl.decls is not None:
+            fields = OrderedDict((field.name, _decl_to_type(field.type, extra_types)) for field in decl.decls)
+        else:
+            fields = OrderedDict()
+
         if decl.name is not None:
             key = 'struct ' + decl.name
-            if key in extra_types:
-                struct = extra_types[key]
-            else:
-                extra_types[key] = struct
+            struct = extra_types.get(key, None)
+            if struct is None:
+                struct = ALL_TYPES.get(key, None)
 
-        if decl.decls is not None:
-            for field in decl.decls:
-                struct.fields[field.name] = _decl_to_type(field.type, extra_types)
+            if struct is None:
+                struct = SimStruct(fields, decl.name)
+            elif not struct.fields:
+                struct.fields = fields
+            elif fields and struct.fields != fields:
+                raise ValueError("Redefining body of " + key)
+
+            extra_types[key] = struct
+        else:
+            struct = SimStruct(fields)
         return struct
 
     elif isinstance(decl, pycparser.c_ast.Union):
-        members = {child[1].name: _decl_to_type(child[1].type, extra_types) for child in decl.children()}
-        return SimUnion(members)
+        if decl.decls is not None:
+            fields = {field.name: _decl_to_type(field.type, extra_types) for field in decl.decls}
+        else:
+            fields = {}
+
+        if decl.name is not None:
+            key = 'union ' + decl.name
+            if key in extra_types:
+                union = extra_types[key]
+            elif key in ALL_TYPES:
+                union = ALL_TYPES[key]
+            else:
+                union = None
+
+            if union is None:
+                union = SimUnion(fields, decl.name)
+            elif not union.members:
+                union.members = fields
+            elif fields and union.members != fields:
+                raise ValueError("Redefining body of " + key)
+
+            extra_types[key] = union
+        else:
+            union = SimUnion(fields)
+        return union
 
     elif isinstance(decl, pycparser.c_ast.IdentifierType):
         key = ' '.join(decl.names)
@@ -1154,6 +1398,8 @@ def _parse_const(c):
     else:
         raise ValueError(c)
 
+if pycparser is not None:
+    _accepts_scope_stack()
 
 try:
     register_types(parse_types("""
