@@ -1,8 +1,9 @@
+#define __STDC_FORMAT_MACROS 1
 #include <unicorn/unicorn.h>
 
+#include <cinttypes>
 #include <cstring>
 #include <cstdint>
-#include <cinttypes>
 
 #include <memory>
 #include <map>
@@ -10,8 +11,10 @@
 #include <unordered_set>
 #include <unordered_map>
 #include <set>
+#include <algorithm>
 
 extern "C" {
+#include <assert.h>
 #include <libvex.h>
 #include <pyvex.h>
 }
@@ -47,6 +50,7 @@ typedef enum stop {
 	STOP_SEGFAULT,
 	STOP_ZERO_DIV,
 	STOP_NODECODE,
+	STOP_HLT,
 } stop_t;
 
 typedef struct block_entry {
@@ -232,17 +236,28 @@ public:
 		executed_pages.clear();
 
 		// error if pc is 0
+		// TODO: why is this check here and not elsewhere
 		if (pc == 0) {
 			stop_reason = STOP_ZEROPAGE;
+			cur_steps = 0;
 			return UC_ERR_MAP;
 		}
 
 		uc_err out = uc_emu_start(uc, pc, 0, 0, 0);
+		if (out == UC_ERR_OK && stop_reason == STOP_NOSTART && get_instruction_pointer() == 0) {
+		    // handle edge case where we stop because we reached our bogus stop address (0)
+		    commit();
+		    stop_reason = STOP_ZEROPAGE;
+		}
 		rollback();
 
 		if (out == UC_ERR_INSN_INVALID) {
 			stop_reason = STOP_NODECODE;
 		}
+
+		// if we errored out right away, fix the step count to 0
+		if (cur_steps == -1) cur_steps = 0;
+
 		return out;
 	}
 
@@ -293,9 +308,6 @@ public:
 		stop_reason = reason;
 		//LOG_D("stop: %s", msg);
 		uc_emu_stop(uc);
-
-		// if we errored out right away, fix the step count to 0
-		if (cur_steps == -1) cur_steps = 0;
 	}
 
 	void step(uint64_t current_address, int32_t size, bool check_stop_points=true) {
@@ -458,8 +470,9 @@ public:
 				memset(bitmap, TAINT_NONE, sizeof(PageBitmap));
 			}
 		} else {
-			if (address == 0x1000) {
-				printf("[sim_unicorn] You've mapped something at 0x1000! "
+		    // TODO: un-hardcode this address, or at least do this warning from python land
+			if (address == 0x4000) {
+				printf("[sim_unicorn] You've mapped something at 0x4000! "
 					"Please don't do that, I put my GDT there!\n");
 			} else {
 				printf("[sim_unicorn] Something very bad is happening; please investigate. "
@@ -527,35 +540,21 @@ public:
 
 	std::pair<uint64_t, size_t> cache_page(uint64_t address, size_t size, char* bytes, uint64_t permissions)
 	{
-		//printf("caching page %#lx - %#lx.\n", address, address + size);
-		// Make sure this page is not overlapping with any existing cached page
-		auto after = page_cache->lower_bound(address);
-		auto before = page_cache->lower_bound(address);
+		assert(address % 0x1000 == 0);
+		assert(size % 0x1000 == 0);
 
-		if (after != page_cache->end()) {
-			if (address + size >= after->first) {
-				if (address >= after->first) {
-					printf("[%#" PRIx64 ", %#" PRIx64 "](%#zx) overlaps with [%#" PRIx64 ", %#" PRIx64 "](%#zx).\n", address, address + size, size, after->first, after->first + after->second.size, after->second.size);
-					// A complete overlap
-					return std::make_pair(address, size);
-				}
-				size = after->first - address;
+		for (uint64_t offset = 0; offset < size; offset += 0x1000)
+		{
+			auto page = page_cache->find(address+offset);
+			if (page != page_cache->end())
+			{
+				fprintf(stderr, "[%#" PRIx64 ", %#" PRIx64 "](%#zx) already in cache.\n", address+offset, address+offset + 0x1000, 0x1000);
+				assert(page->second.size == 0x1000);
+				assert(memcmp(page->second.bytes, bytes + offset, 0x1000) == 0);
+
+				continue;
 			}
-		}
-		if (before != page_cache->begin()) {
-			before--;
-			if (address < before->first + before->second.size) {
-				if (address + size <= before->first + before->second.size) {
-					// A complete overlap
-					printf("[%#" PRIx64 ", %#" PRIx64 "](%#zx) overlaps with [%#" PRIx64 ", %#" PRIx64 "](%#zx).\n", address, address + size, size, after->first, after->first + after->second.size, after->second.size);
-					return std::make_pair(address, size);
-				}
-				size = address + size - (before->first + before->second.size);
-				address = before->first + before->second.size;
-			}
-		}
 
-		for (uint64_t offset = 0; offset < size; offset += 0x1000) {
 			uint8_t *copy = (uint8_t *)malloc(0x1000);
 			CachedPage cached_page = {
 				0x1000,
@@ -569,66 +568,71 @@ public:
 		return std::make_pair(address, size);
 	}
 
-	void uncache_page(uint64_t address) {
-		if ((address & 0xfff) != 0) {
-			printf("Warning: Address #%" PRIx64 " passed to uncache_page is not aligned\n", address);
-			return;
-		}
-
+    void wipe_page_from_cache(uint64_t address) {
 		auto page = page_cache->find(address);
 		if (page != page_cache->end()) {
 			//printf("Internal: unmapping %#llx size %#x, result %#x", page->first, page->second.size, uc_mem_unmap(uc, page->first, page->second.size));
-			uc_mem_unmap(uc, page->first, page->second.size);
-			//free(page->second.bytes); // other forks might need this I guess :( how the HELL do you make memory management sane?
+			uc_err err = uc_mem_unmap(uc, page->first, page->second.size);
+			//if (err) {
+			//	fprintf(stderr, "wipe_page_from_cache [%#lx, %#lx]: %s\n", page->first, page->first + page->second.size, uc_strerror(err));
+			//}
+			free(page->second.bytes); // might explode
 			page_cache->erase(page);
 		} else {
 			//printf("Uh oh! Couldn't find page at %#llx\n", address);
 		}
-	}
+    }
+
+    void uncache_pages_touching_region(uint64_t address, uint64_t length)
+    {
+    	    address &= ~(0x1000-1);
+
+	    for (uint64_t offset = 0; offset < length; offset += 0x1000)
+	    {
+            	    wipe_page_from_cache(address + offset);
+	    }
+
+    }
+
+    void clear_page_cache()
+    {
+        while (!page_cache->empty())
+        {
+            wipe_page_from_cache(page_cache->begin()->first);
+        }
+    }
 
 	bool map_cache(uint64_t address, size_t size) {
-		auto it = page_cache->lower_bound(address);
+		assert(address % 0x1000 == 0);
+		assert(size % 0x1000 == 0);
 
-		if (it == page_cache->end() && it != page_cache->begin()) {
-			// Maybe the previous one works?
-			it--;
-		}
+		bool success = true;
 
-		if (it != page_cache->end()) {
-			uint64_t cached_page_addr = it->first;
-			if (cached_page_addr > address && it != page_cache->begin()) {
-				it--;
-				cached_page_addr = it->first;
+		for (uint64_t offset = 0; offset < size; offset += 0x1000)
+		{
+			auto page = page_cache->find(address+offset);
+			if (page == page_cache->end())
+			{
+				success = false;
+				continue;
 			}
-			auto itt = it->second;
-			size_t page_size = itt.size;
-			uint8_t *bytes = itt.bytes;
-			uint64_t permissions = itt.perms;
 
-			if (address >= cached_page_addr && address < cached_page_addr + page_size) {
-				while (cached_page_addr < address + size) {
-					//LOG_D("hit cache [%#lx, %#lx]", address, address + size);
-					uc_err err = uc_mem_map_ptr(uc, cached_page_addr, page_size, permissions, bytes);
-					if (err) {
-						//LOG_E("map_cache [%#lx, %#lx]: %s", address, address + size, uc_strerror(err));
-						return false;
-					}
+			auto cached_page = page->second;
+			size_t page_size = cached_page.size;
+			uint8_t *bytes = cached_page.bytes;
+			uint64_t permissions = cached_page.perms;
 
-					it++;
-					if (it != page_cache->end()) {
-						cached_page_addr = it->first;
-						page_size = it->second.size;
-						bytes = it->second.bytes;
-						permissions = it->second.perms;
-					} else {
-						break;
-					}
-				}
-				return true;
+			assert(page_size == 0x1000);
+
+			//LOG_D("hit cache [%#lx, %#lx]", address, address + size);
+			uc_err err = uc_mem_map_ptr(uc, page->first, page_size, permissions, bytes);
+			if (err) {
+				fprintf(stderr, "map_cache [%#lx, %#lx]: %s\n", address, address + size, uc_strerror(err));
+				success = false;
+				continue;
 			}
 		}
-		//LOG_D("cache miss.");
-		return false;
+		return success;
 	}
 
 	bool in_cache(uint64_t address) {
@@ -834,6 +838,7 @@ public:
 			used_registers = &search->second.used_registers;
 		} else {
 			// wtf i hate c++...
+			VexRegisterUpdates pxControl = VexRegUpdUnwindregsAtMemAccess;
 			auto& entry = this->block_cache->emplace(std::make_pair(address, block_entry_t())).first->second;
 			entry.try_unicorn = true;
 			clobbered_registers = &entry.clobbered_registers;
@@ -842,7 +847,8 @@ public:
 			std::unique_ptr<uint8_t[]> instructions(new uint8_t[size]);
 			uc_mem_read(this->uc, address, instructions.get(), size);
 			VEXLiftResult *lift_ret = vex_lift(
-					this->vex_guest, this->vex_archinfo, instructions.get(), address, 99, size, 1, 0, 0, 1, 0
+					this->vex_guest, this->vex_archinfo, instructions.get(), address, 99, size, 1, 0, 0, 1, 0,
+					pxControl
 					);
 
 
@@ -1410,8 +1416,13 @@ bool simunicorn_cache_page(State *state, uint64_t address, uint64_t length, char
 }
 
 extern "C"
-void simunicorn_uncache_page(State *state, uint64_t address) {
-	state->uncache_page(address);
+void simunicorn_uncache_pages_touching_region(State *state, uint64_t address, uint64_t length) {
+	state->uncache_pages_touching_region(address, length);
+}
+
+extern "C"
+void simunicorn_clear_page_cache(State *state) {
+	state->clear_page_cache();
 }
 
 // Tracking settings
@@ -1420,3 +1431,8 @@ void simunicorn_set_tracking(State *state, bool track_bbls, bool track_stack) {
 	state->track_bbls = track_bbls;
 	state->track_stack = track_stack;
 }
+
+extern "C"
+bool simunicorn_in_cache(State *state, uint64_t address) {
+	return state->in_cache(address);
+}
